"""
Example of defining a custom (image) transform using FFCV.
For tutorial, see https://docs.ffcv.io/ffcv_examples/transform_with_inds.html.

"""
from dataclasses import replace
import time
from typing import Callable, Optional, Tuple
import numpy as np
import torchvision

from ffcv.fields import IntField, RGBImageField
from ffcv.fields.decoders import IntDecoder
from ffcv.loader import Loader, OrderOption
from ffcv.pipeline.compiler import Compiler
from ffcv.pipeline.operation import Operation, AllocationQuery
from ffcv.pipeline.state import State
from ffcv.transforms import ToTensor
from ffcv.writer import DatasetWriter


class CorruptFixedLabels(Operation):
    def generate_code(self) -> Callable:
        # dst will be None since we don't ask for an allocation
        parallel_range = Compiler.get_iterator()
        def corrupt_fixed(labs, _, inds):
            for i in parallel_range(labs.shape[0]):
                # Because the random seed is tied to the image index, the
                # same images will be corrupted every epoch:
                np.random.seed(inds[i])
                if np.random.rand() < 0.2:
                    # They will also be corrupted to a deterministic label:
                    labs[i] = np.random.randint(low=0, high=10)
            return labs

        corrupt_fixed.is_parallel = True
        corrupt_fixed.with_indices = True
        return corrupt_fixed

    def declare_state_and_memory(self, previous_state: State) -> Tuple[State, Optional[AllocationQuery]]:
        # No updates to state or extra memory necessary!
        return previous_state, None

# Step 1: Create an FFCV-compatible CIFAR-10 dataset
ds = torchvision.datasets.CIFAR10('/tmp', train=True, download=True)
writer = DatasetWriter('/tmp/cifar.beton', {
    'image': RGBImageField(),
    'label': IntField()
})
writer.from_indexed_dataset(ds)

# Step 2: Create data loaders
BATCH_SIZE = 512
label_pipelines = {
    'with': [IntDecoder(), CorruptFixedLabels(), ToTensor()],
    'without': [IntDecoder(), ToTensor()]
}

for name, pipeline in label_pipelines.items():
    # Use SEQUENTIAL ordering to compare labels.
    loader = Loader(f'/tmp/cifar.beton', batch_size=BATCH_SIZE,
                    num_workers=8, order=OrderOption.SEQUENTIAL,
                    drop_last=True, pipelines={'label': pipeline})

    # First epoch includes compilation time
    for ims, labs in loader: pass
    start_time = time.time()
    for ep in range(20):
        for i, (ims, labs) in enumerate(loader):
            if i == 0: # Inspect first batch
                print(f'> Labels (epoch {ep:2}): {labs[:40,0].tolist()}')
    print(f'Method: {name} | Shape: {ims.shape} | Time per epoch: {(time.time() - start_time) / 100:.5f}s')


"""
Random resized crop, similar to torchvision.transforms.RandomResizedCrop
"""
from dataclasses import replace
from .utils import fast_crop
import numpy as np
from typing import Callable, Optional, Tuple
from ..pipeline.allocation_query import AllocationQuery
from ..pipeline.operation import Operation
from ..pipeline.state import State
from ..pipeline.compiler import Compiler

class RandomResizedCrop(Operation):
    """Crop a random portion of image with random aspect ratio and resize it to
    a given size. Chances are you do not want to use this augmentation and
    instead want to include RRC as part of the decoder, by using the 
    :cla:`~ffcv.fields.rgb_image.ResizedCropRGBImageDecoder` class.

    Parameters
    ----------
    scale : Tuple[float, float]
        Lower and upper bounds for the ratio of random area of the crop.
    ratio : Tuple[float, float]
        Lower and upper bounds for random aspect ratio of the crop.
    size : int
        Side length of the output.
    """
    def __init__(self, scale: Tuple[float, float], ratio: Tuple[float, float], size: int):
        super().__init__()
        self.scale = scale
        self.ratio = ratio
        self.size = size
        

    def generate_code(self) -> Callable:
        scale, ratio = self.scale, self.ratio
        if isinstance(scale, tuple):
            scale = np.array(scale)
        if isinstance(ratio, tuple):
            ratio = np.array(ratio)
        my_range = Compiler.get_iterator()
        def random_resized_crop(images, dst):
            for idx in my_range(images.shape[0]):
                i, j, h, w = fast_crop.get_random_crop(images[idx].shape[0],
                                                    images[idx].shape[1],
                                                    scale,
                                                    ratio)
                fast_crop.resize_crop(images[idx], i, i + h, j, j + w, dst[idx])
            return dst

        random_resized_crop.is_parallel = True
        return random_resized_crop

    def declare_state_and_memory(self, previous_state: State) -> Tuple[State, Optional[AllocationQuery]]:
        return replace(previous_state, jit_mode=True, shape=(self.size, self.size, 3)), \
               AllocationQuery((self.size, self.size, 3), dtype=previous_state.dtype)





"""
Cutout augmentation (https://arxiv.org/abs/1708.04552)
"""
import numpy as np
from typing import Callable, Optional, Tuple
from dataclasses import replace

from ffcv.pipeline.compiler import Compiler
from ..pipeline.allocation_query import AllocationQuery
from ..pipeline.operation import Operation
from ..pipeline.state import State

class Cutout(Operation):
    """Cutout data augmentation (https://arxiv.org/abs/1708.04552).

    Parameters
    ----------
    crop_size : int
        Size of the random square to cut out.
    fill : Tuple[int, int, int], optional
        An RGB color ((0, 0, 0) by default) to fill the cutout square with.
        Useful for when a normalization layer follows cutout, in which case
        you can set the fill such that the square is zero
        post-normalization.
    """
    def __init__(self, crop_size: int, fill: Tuple[int, int, int] = (0, 0, 0)):
        super().__init__()
        self.crop_size = crop_size
        self.fill = np.array(fill)

    def generate_code(self) -> Callable:
        my_range = Compiler.get_iterator()
        crop_size = self.crop_size
        fill = self.fill

        def cutout_square(images, *_):
            for i in my_range(images.shape[0]):
                # Generate random origin
                coord = (
                    np.random.randint(images.shape[1] - crop_size + 1),
                    np.random.randint(images.shape[2] - crop_size + 1),
                )
                # Black out image in-place
                images[i, coord[0]:coord[0] + crop_size, coord[1]:coord[1] + crop_size] = fill

            return images
        cutout_square.is_parallel = True

        return cutout_square

    def declare_state_and_memory(self, previous_state: State) -> Tuple[State, Optional[AllocationQuery]]:
        return replace(previous_state, jit_mode=True), None



from collections import defaultdict

import numpy as np

from ..base import MemoryManager, MemoryContext
from ..common import BATCHES_TYPE
from .schedule import Schedule, ScheduleExecutor, compute_schedule


class ProcessCacheContext(MemoryContext):

    def __init__(self, manager: MemoryManager, batches: BATCHES_TYPE):
        self.manager = manager
        self.fname = manager.reader.file_name
        self.batches = batches
        self.page_size = manager.reader.page_size

    @property
    def state(self):
        return (self.memory, self.manager.ptrs,
                self.manager.sizes, self.page_to_slot)

    def __enter__(self):
        pages_at_batch = []
        for batch in self.batches:
            pages_needed = set()
            for sample_id in batch:
                pages_needed.update(self.manager.sample_to_pages[sample_id])
            pages_at_batch.append(pages_needed)

        self.schedule = compute_schedule(pages_at_batch)
        self.memory = np.zeros((self.schedule.num_slots, self.page_size),
                               dtype='<u1')
        self.executor = ScheduleExecutor(self.fname,
                                         self.schedule,
                                         self.memory)

        try:
            max_page = max(self.schedule.page_to_slot.keys())
        except ValueError:
            max_page = -1

        # We need max + 1 slots
        # We use a table as it's O(1) indexing. Pages for the header will
        # be unused however so we are losing some space
        self.page_to_slot = np.zeros(max_page + np.uint32(1), dtype=np.uint32)

        for p, s in self.schedule.page_to_slot.items():
            self.page_to_slot[p] = s

        self.executor.__enter__()

    def start_batch(self, batch: int):
        self.executor.load_batch(batch)
        return super().start_batch(batch)


    def __exit__(self, *args):
        self.executor.__exit__(*args)




from dataclasses import dataclass
from typing import Literal, Tuple

import torch as ch
import numpy as np

@dataclass
class State:
    jit_mode: bool
    device: ch.device
    shape: Tuple[int, ...]
    dtype: np.dtype
    
    # Assess the validity of a pipeline stage
    def __post_init__(self):
        if self.jit_mode and self.device != ch.device('cpu'):
            raise AssertionError("Can't be in JIT mode and on the GPU")
        if self.jit_mode and isinstance(self.dtype, ch.dtype):
            raise AssertionError("Can't allocate a torch tensor in JIT mode")


"""
Random horizontal flip
"""
from dataclasses import replace
from numpy.random import rand
from typing import Callable, Optional, Tuple
from ..pipeline.allocation_query import AllocationQuery
from ..pipeline.operation import Operation
from ..pipeline.state import State
from ..pipeline.compiler import Compiler

class RandomHorizontalFlip(Operation):
    """Flip the image horizontally with probability flip_prob.
    Operates on raw arrays (not tensors).

    Parameters
    ----------
    flip_prob : float
        The probability with which to flip each image in the batch
        horizontally.
    """

    def __init__(self, flip_prob: float = 0.5):
        super().__init__()
        self.flip_prob = flip_prob

    def generate_code(self) -> Callable:
        my_range = Compiler.get_iterator()
        flip_prob = self.flip_prob

        def flip(images, dst):
            should_flip = rand(images.shape[0]) < flip_prob
            for i in my_range(images.shape[0]):
                if should_flip[i]:
                    dst[i] = images[i, :, ::-1]
                else:
                    dst[i] = images[i]

            return dst

        flip.is_parallel = True
        return flip

    def declare_state_and_memory(self, previous_state: State) -> Tuple[State, Optional[AllocationQuery]]:
        return (replace(previous_state, jit_mode=True), 
                AllocationQuery(previous_state.shape, previous_state.dtype))



import pytest
import numpy as np
from uuid import uuid4
from ffcv.fields.ndarray import NDArrayField, NDArrayDecoder
from ffcv.writer import DatasetWriter
from ffcv.loader import Loader, OrderOption
from tempfile import NamedTemporaryFile

class StringDecoder(NDArrayDecoder):
    pass

class StringField(NDArrayField):
    def __init__(self, max_len: int, pad_char='\0'):
        self.max_len = max_len
        self.pad_char = pad_char
        super().__init__(np.dtype('uint8'), (max_len,))
    
    def encode(self, destination, field, malloc):
        padded_field = (field + self.pad_char * self.max_len)[:self.max_len]
        field = np.frombuffer(padded_field.encode('ascii'), dtype='uint8')
        return super().encode(destination, field, malloc)

MAX_STRING_SIZE = 100

class CaptionDataset:
    def __init__(self, N):
        self.captions = [str(uuid4())[:np.random.randint(50)] for _ in range(N)]

    def __getitem__(self, idx):
        return (self.captions[idx],)

    def __len__(self):
        return len(self.captions)

def test_string_field():
    dataset = CaptionDataset(100)

    with NamedTemporaryFile() as handle:
        writer = DatasetWriter(handle.name, {
            'label': StringField(MAX_STRING_SIZE)
        })

        writer.from_indexed_dataset(dataset)
        loader = Loader(handle.name,
                    batch_size=10,
                    num_workers=2,
                    order=OrderOption.RANDOM,
                    pipelines={
                        'label': [StringDecoder()]
                    },
                    custom_fields={
                        'label': StringField
                    })
        
        all_caps = []
        for x, in loader:
            for cap in x:
                all_caps.append(cap.tobytes().decode('ascii').replace('\0', ''))
        assert set(all_caps) == set(dataset.captions)

def test_no_custom_field():
    dataset = CaptionDataset(100)

    with NamedTemporaryFile() as handle:
        writer = DatasetWriter(handle.name, {
            'label': StringField(MAX_STRING_SIZE)
        })

        writer.from_indexed_dataset(dataset)
        with pytest.raises(ValueError):
            Loader(handle.name,
                    batch_size=10,
                    num_workers=2,
                    order=OrderOption.RANDOM,
                    pipelines={
                        'label': [StringDecoder()]
                    })
        


import numpy as np
from tqdm import tqdm
from assertpy import assert_that
from torch.utils.data import Dataset
import logging
from time import time
import os
from assertpy import assert_that
from tempfile import NamedTemporaryFile

from ffcv.writer import DatasetWriter
from ffcv.reader import Reader
from ffcv.fields import IntField, RGBImageField
from ffcv.pipeline.compiler import Compiler
from ffcv.memory_managers import OSCacheManager

class DummyDataset(Dataset):

    def __init__(self, length):
        self.length = length
        
    def __len__(self):
        return self.length

    def __getitem__(self, index):
        if index >= self.length:
            raise IndexError
        
        np.random.seed(37 + index)
        dims = tuple([128, 128, 3])
        image_data = np.random.randint(low=0, high=255, size=dims, dtype='uint8')
        return index, image_data



def create_and_validate(length, mode='raw', compile=False):

    dataset = DummyDataset(length)

    with NamedTemporaryFile() as handle:
        name = handle.name
        writer = DatasetWriter(name, {
            'index': IntField(),
            'value': RGBImageField(write_mode=mode)
        }, num_workers=2)

        writer.from_indexed_dataset(dataset, chunksize=5)
            
        reader = Reader(name)
        manager = OSCacheManager(reader)
        context = manager.schedule_epoch(np.array([0, 1]))
        
        Compiler.set_enabled(compile)

        with context:
            Decoder = RGBImageField().get_decoder_class()
            decoder = Decoder()
            decoder.accept_globals(reader.metadata['f1'], manager.compile_reader())

        decode = Compiler.compile(decoder.generate_code())

        assert_that(reader.metadata).is_length(length)
        buff = np.zeros((1, 128, 128, 3), dtype='uint8')
        
        for i in range(length):
            result = decode(np.array([i]), buff, reader.metadata['f1'], context.state)[0]
            _, ref_image = dataset[i]
            assert_that(result.shape).is_equal_to(ref_image.shape)
            if mode == 'jpg':
                dist = np.abs(ref_image.astype('float') - result.astype('float'))
                assert_that(dist.mean()).is_less_than(80)
            else:
                assert_that(np.all(ref_image == result)).is_true()
                
def test_simple_image_dataset_raw():
    create_and_validate(500, 'raw')

def test_simple_image_dataset_jpg():
    create_and_validate(100, 'jpg')

def test_simple_image_dataset_raw_compile():
    create_and_validate(500, 'raw', True)

def test_simple_image_dataset_jpg_compile():
    create_and_validate(100, 'jpg', True)




import numpy as np
from time import sleep
from os import SEEK_END
from multiprocessing import Value
from .utils import align_to_page
import ctypes

class MemoryAllocator():
    def __init__(self, fname, offset_start, page_size):
        self.fname = fname
        self.offset = align_to_page(offset_start, page_size)
        self.next_page_allocated  = Value(ctypes.c_uint64, 0)
        self.next_page_written  = Value(ctypes.c_uint64, 0)

        self.page_size = page_size
        self.page_offset = 0
        self.my_page = -1

        self.page_data = np.zeros(self.page_size, '<u1')
        self.allocations = []
        self.current_sample_id = None

    def __enter__(self):
        self.fp = open(self.fname, 'ab', buffering=0)

    def set_current_sample(self, current_sample_id):
        self.current_sample_id = current_sample_id

    @property
    def space_left_in_page(self):
        # We don't have a page allocated yet
        if self.my_page < 0:
            return 0
        return self.page_size - self.page_offset

    def malloc(self, size):
        # print(f"Allocating {size} bytes")
        if size > self.page_size:
            raise ValueError(f"Tried allocating {size} but" +
                             f" page size is {self.page_size}")

        if size > self.space_left_in_page:
            self.flush_page()
            # We book the next available page in the file
            with self.next_page_allocated.get_lock():
                self.my_page = self.next_page_allocated.value
                self.next_page_allocated.value = self.my_page + 1

            self.page_offset = 0
            # This is a new page so we erate the content of the buffer
            self.page_data.fill(0)

            # We check if we already allocated space for this sample on
            # the page that is now full
            region_in_previous_page = False
            while self.allocations and self.allocations[-1][0] == self.current_sample_id:
                # We have to revert the allocations we did and we are giving
                # up on this sample.
                self.allocations.pop()
                # We found at least memory region from the preivous page
                region_in_previous_page = True

            # The writer will restart from this freshly allocated page
            if region_in_previous_page:
                raise MemoryError("Not enough memory to fit the whole sample")

        previous_offset = self.page_offset
        self.page_offset += size

        buffer = self.page_data[previous_offset:self.page_offset]
        ptr = self.offset + self.my_page * self.page_size + previous_offset

        # We return the pointer to the location in file and where to write
        # the data
        self.allocations.append((self.current_sample_id, ptr, size))
        return ptr, buffer

    def flush_page(self):
        # If we haven't allocated any page we end there
        if self.my_page < 0:
            return

        # We shouldn't have allocated a page and have nothing to write on it
        assert self.page_offset != 0
        # Wait until it's my turn to write
        while self.next_page_written.value != self.my_page:
            # Essentially a spin lock
            # TODO we could replace it with like exponential backoff
            sleep(0.001)
            pass

        # Now it's my turn to write

        expected_file_offset = self.offset + self.my_page * self.page_size
        # in order to be aligned with page size
        # If this is the first page we have to pad with zeros
        if self.my_page == 0:
            # print("Padding headers to align with page size")
            current_location = self.fp.seek(0, SEEK_END)
            null_bytes_to_write = expected_file_offset - current_location
            self.fp.write(np.zeros(null_bytes_to_write, dtype='<u1').tobytes())
            # print(f"current file pointer is no {self.fp.tell()} and should be {expected_file_offset}")


        self.fp.seek(expected_file_offset)

        # print(f"Writing page {self.my_page} at offset {self.fp.tell()}")
        self.fp.write(self.page_data.tobytes())
        # print(f"Done writing {self.my_page} at offset {self.fp.tell()}")

        # We warn other processes that they are free to write the next page
        with self.next_page_written.get_lock():
            self.next_page_written.value += 1



    # Flush the last page and
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.flush_page()
        self.fp.close()



import os
from tempfile import NamedTemporaryFile
from time import sleep, time

import numpy as np
from tqdm import tqdm
from assertpy import assert_that
from torch.utils.data import Dataset

from ffcv.writer import DatasetWriter
from ffcv.reader import Reader
from ffcv.fields import BytesField, IntField
from ffcv.pipeline.compiler import Compiler
from ffcv.memory_managers import OSCacheManager
from ffcv.libffcv import memcpy

from ..decorator import benchmark
from ..benchmark import Benchmark

class DummyDataset(Dataset):

    def __init__(self, l, size):
        self.l = l
        self.size = size

    def __len__(self):
        return self.l

    def __getitem__(self, index):
        if index > self.l:
            raise IndexError
        np.random.seed(index)
        return index, np.random.randint(0, 255, size=self.size, dtype='u1')

@benchmark({
    'num_samples': [3000],
    'size_bytes': [
        32 * 32 * 3, # CIFAR RAW image size,
        500 * 300 * 3, # IMAGENET raw image size,
        128 * 1024, # IMAGENET jpg image size,
    ],
    'compiled': [
        True
    ],
    'random_reads': [True, False],
    'n': [3000]
})
class MemoryReadBytesBench(Benchmark):

    def __init__(self, num_samples, size_bytes, random_reads, n, compiled):
        self.num_samples = num_samples
        self.size_bytes = size_bytes
        self.random_reads = random_reads
        self.n = n
        self.compiled = compiled
        
    def __enter__(self):
        self.handle = NamedTemporaryFile()
        handle = self.handle.__enter__()
        name = handle.name
        dataset = DummyDataset(self.num_samples, self.size_bytes)
        writer = DatasetWriter(self.num_samples, name, {
            'index': IntField(),
            'value': BytesField()
        })

        with writer:
            writer.write_pytorch_dataset(dataset, num_workers=-1, chunksize=100)

        reader = Reader(name)
        manager = OSCacheManager(reader)
        context = manager.schedule_epoch(np.arange(self.num_samples))
        context.__enter__()
        self.context = context

        Compiler.set_enabled(self.compiled)
        memcpy_c = Compiler.compile(memcpy)

        read_fn = manager.compile_reader()

        if self.random_reads:
            indices = np.random.choice(self.num_samples, self.n, replace=False)
        else:
            indices = np.arange(self.num_samples)[:self.n]

        addresses = reader.alloc_table['ptr'][indices]
        
        self.buffer = np.zeros(self.size_bytes, dtype='<u1')
        
        def code(buff, state):
            for i in range(addresses.shape[0]):
                memcpy_c(read_fn(addresses[i], state), buff)
        
        self.code = Compiler.compile(code)

    
    def run(self):
        self.code(self.buffer, self.context.state)
        
    def __exit__(self, *args):
        self.handle.__exit__(*args)


from typing import List

import numpy as np


from .fields.base import Field
from .fields import (
    FloatField, IntField, RGBImageField,
    BytesField, NDArrayField, JSONField,
    TorchTensorField
)

CURRENT_VERSION = 2

# Note that in this file we use dtypes in the format <u4 indead of uint32. This
# forces endinaness of the data making datasets compatible between different
# CPU architectures ARM for example is big-endian and x86 is little endian. We
# fix the endianness to little-endian as we suspect that this library will
# mostly be used on x86 architecture (at least in the near future)

# Type describing the data coming
HeaderType = np.dtype([
    ('version', '<u2'),
    ('num_fields', '<u2'),
    ('page_size', '<u4'),
    ('num_samples', '<u8'),
    ('alloc_table_ptr', '<u8')
], align=True)

ALLOC_TABLE_TYPE = np.dtype([
    ('sample_id', '<u8'),
    ('ptr', '<u8'),
    ('size', '<u8'),
])

FieldDescType = np.dtype([
    # This identifier will inform us on how to decode that field
    ('type_id', '<u1'),
    ('name', ('<u1', 16)),
    # Data that will depend on the type of the field (some might need arguments
    # like images, but some might not like integers and floats)
    ('arguments', ('<u1', (1024, )))
], align=True)

# Map from type_id to the handler for that kind of data
TYPE_ID_HANDLER = {
    255 : None,
    0   : FloatField,
    1   : IntField,
    2   : RGBImageField,
    3   : BytesField,
    4   : NDArrayField,
    5   : JSONField,
    6   : TorchTensorField
}

# Parse the fields descriptors from the header of the dataset
# Return the corresponding handlers

def get_handlers(field_descriptors):
    handlers = []
    for field_descriptor in field_descriptors:
        type_id = field_descriptor['type_id']
        Handler = TYPE_ID_HANDLER[type_id]
        if Handler is None:
            handlers.append(None)
        else:
            handlers.append(Handler.from_binary(field_descriptor['arguments']))
    return handlers



# From a list of handlers return the combined data type that will
# describe a complete sample
def get_metadata_type(handlers: List[Field]) -> np.dtype:
    return np.dtype([('', handler.metadata_type) for handler in handlers],
                    align=True)



from functools import partial
from typing import Callable, List, Mapping
from os import SEEK_END, path, sched_getaffinity
import numpy as np
from time import sleep
import ctypes
from multiprocessing import (shared_memory, cpu_count, Queue, Process, Value)

from tqdm import tqdm
from tqdm.contrib.concurrent import thread_map

from .utils import chunks, is_power_of_2
from .fields.base import Field
from .memory_allocator import MemoryAllocator
from .types import (TYPE_ID_HANDLER, get_metadata_type, HeaderType,
                    FieldDescType, CURRENT_VERSION, ALLOC_TABLE_TYPE)


MIN_PAGE_SIZE = 1 << 21  # 2MiB, which is the most common HugePage size
MAX_PAGE_SIZE = 1 << 32  # Biggest page size that will not overflow uint32

def from_shard(shard, pipeline):
    # We import webdataset here so that it desn't crash if it's not required
    # (Webdataset is an optional depdency)
    from webdataset import WebDataset

    dataset = WebDataset(shard)
    dataset = pipeline(dataset)
    return dataset

def count_samples_in_shard(shard, pipeline):
    #
    # We count the length of the dataset
    # We are not using __len__ since it might not be implemented
    count = 0
    for _ in from_shard(shard, pipeline):
        count += 1

    return count


def handle_sample(sample, dest_ix, field_names, metadata, allocator, fields):
    # We should only have to retry at least one
    for i in range(2):
        try:
            allocator.set_current_sample(dest_ix)
            # We extract the sample in question from the dataset
            # We write each field individually to the metadata region
            for field_name, field, field_value in zip(field_names, fields.values(), sample):
                destination = metadata[field_name][dest_ix: dest_ix + 1]
                field.encode(destination, field_value, allocator.malloc)
            # We managed to write all the data without reaching
            # the end of the page so we stop retrying
            break
        # If we couldn't fit this sample in the previous page we retry once from a fresh page
        except MemoryError:
            # We raise the error if it happens on the second try
            if i == 1:
                raise

def worker_job_webdataset(input_queue, metadata_sm, metadata_type, fields,
               allocator, done_number, allocations_queue, pipeline):

    metadata = np.frombuffer(metadata_sm.buf, dtype=metadata_type)
    field_names = metadata_type.names

    # This `with` block ensures that all the pages allocated have been written
    # onto the file
    with allocator:
        while True:
            todo = input_queue.get()

            if todo is None:
                # No more work left to do
                break

            shard, offset = todo

            # For each sample in the chunk
            done = 0
            for i, sample in enumerate(from_shard(shard, pipeline)):
                done += 1
                dest_ix = offset + i
                handle_sample(sample, dest_ix, field_names, metadata, allocator, fields)

            # We warn the main thread of our progress
            with done_number.get_lock():
                done_number.value += done

    allocations_queue.put(allocator.allocations)



def worker_job_indexed_dataset(input_queue, metadata_sm, metadata_type, fields,
               allocator, done_number, allocations_queue, dataset):

    metadata = np.frombuffer(metadata_sm.buf, dtype=metadata_type)
    field_names = metadata_type.names

    # This `with` block ensures that all the pages allocated have been written
    # onto the file
    with allocator:
        while True:
            chunk = input_queue.get()

            if chunk is None:
                # No more work left to do
                break

            # For each sample in the chunk
            for dest_ix, source_ix in chunk:
                sample = dataset[source_ix]
                handle_sample(sample, dest_ix, field_names, metadata, allocator, fields)

            # We warn the main thread of our progress
            with done_number.get_lock():
                done_number.value += len(chunk)

    allocations_queue.put(allocator.allocations)


class DatasetWriter():
    """Writes given dataset into FFCV format (.beton).
    Supports indexable objects (e.g., PyTorch Datasets) and webdataset.

    Parameters
    ----------
    fname: str
        File name to store dataset in FFCV format (.beton)
    fields : Mapping[str, Field]
        Map from keys to Field's (order matters!)
    page_size : int
        Page size used internally
    num_workers : int
        Number of processes to use
    """
    def __init__(self, fname: str, fields: Mapping[str, Field],
                 page_size: int = 4 * MIN_PAGE_SIZE, num_workers: int = -1):
        self.fields = fields
        self.fname = fname
        self.metadata_type = get_metadata_type(list(self.fields.values()))

        self.num_workers = num_workers
        # We use all cores by default
        if self.num_workers < 1:
            self.num_workers = len(sched_getaffinity(0))

        if not is_power_of_2(page_size):
            raise ValueError(f'page_size isnt a power of 2')
        if page_size < MIN_PAGE_SIZE:
            raise ValueError(f"page_size can't be lower than{MIN_PAGE_SIZE}")
        if page_size >= MAX_PAGE_SIZE:
            raise ValueError(f"page_size can't be bigger(or =) than{MAX_PAGE_SIZE}")

        self.page_size = page_size

    def prepare(self):

        with open(self.fname, 'wb') as fp:
            # Prepare the header data
            header = np.zeros(1, dtype=HeaderType)[0]
            header['version'] = CURRENT_VERSION
            header['num_samples'] = self.num_samples
            header['num_fields'] = len(self.fields)
            header['page_size'] = self.page_size
            self.header = header

            # We will write the header at the end because we need to know where
            # The memory allocation table is in the file
            # We still write it here to make space for it later
            fp.write(self.header.tobytes())


            # Writes the information about the fields
            fields_descriptor = np.zeros(len(self.fields),
                                              dtype=FieldDescType)
            field_type_to_type_id = {v: k for (k, v) in TYPE_ID_HANDLER.items()}

            fieldname_max_len = fields_descriptor[0]['name'].shape[0]

            for i, (name, field) in enumerate(self.fields.items()):
                type_id = field_type_to_type_id.get(type(field), 255)
                encoded_name = name.encode('ascii')
                encoded_name = np.frombuffer(encoded_name, dtype='<u1')
                actual_length = min(fieldname_max_len, len(encoded_name))
                fields_descriptor[i]['type_id'] = type_id
                fields_descriptor[i]['name'][:actual_length] = (
                    encoded_name[:actual_length])
                fields_descriptor[i]['arguments'][:] = field.to_binary()[0]

            fp.write(fields_descriptor.tobytes())

        total_metadata_size = self.num_samples * self.metadata_type.itemsize

        # Shared memory for all the writers to fill the information
        self.metadata_sm = 3
        self.metadata_start = (HeaderType.itemsize + fields_descriptor.nbytes)

        self.metadata_sm = shared_memory.SharedMemory(create=True,
                                                      size=total_metadata_size)

        self.data_region_start = self.metadata_start + total_metadata_size


    def _write_common(self, num_samples, queue_content, work_fn, extra_worker_args):
        self.num_samples = num_samples

        self.prepare()
        allocation_list = []

        # Makes a memmap to the metadata for the samples

        # We publish all the work that has to be done into a queue
        workqueue: Queue = Queue()
        for todo in queue_content:
            workqueue.put(todo)

        # This will contain all the memory allocations each worker
        # produced. This will go at the end of the file
        allocations_queue: Queue = Queue()

        # We add a token for each worker to warn them that there
        # is no more work to be done
        for _ in range(self.num_workers):
            workqueue.put(None)

        # Define counters we need to orchestrate the workers
        done_number = Value(ctypes.c_uint64, 0)
        allocator = MemoryAllocator(self.fname,
                                    self.data_region_start,
                                    self.page_size)

        # Arguments that have to be passed to the workers
        worker_args = (workqueue, self.metadata_sm,
                       self.metadata_type, self.fields,
                       allocator, done_number,
                       allocations_queue, *extra_worker_args)

        # Create the workers
        processes = [Process(target=work_fn, args=worker_args)
                     for _ in range(self.num_workers)]
        # start the workers
        for p in processes: p.start()
        # Wait for all the workers to be done

        # Display progress
        progress = tqdm(total=self.num_samples)
        previous = 0
        while previous != self.num_samples:
            val = done_number.value
            diff = val - previous
            if diff > 0:
                progress.update(diff)
            previous = val
            sleep(0.1)
        progress.close()

        # Wait for all the workers to be done and get their allocations
        for p in processes:
            content = allocations_queue.get()
            allocation_list.extend(content)

        self.finalize(allocation_list)
        self.metadata_sm.close()
        self.metadata_sm.unlink()


    def from_indexed_dataset(self, dataset,
                              indices: List[int]=None, chunksize=100,
                              shuffle_indices: bool = False):
        """Read dataset from an indexable dataset.
        See https://docs.ffcv.io/writing_datasets.html#indexable-dataset for sample usage.

        Parameters
        ----------
        dataset: Indexable
            An indexable object that implements `__getitem__` and `__len__`.
        indices : List[int]
            Use a subset of the dataset specified by indices.
        chunksize : int
            Size of chunks processed by each worker during conversion.
        shuffle_indices : bool
            Shuffle order of the dataset.
        """
        # If the user didn't specify an order we just add samples
        # sequentially
        if indices is None:
            indices = np.arange(len(dataset))

        if shuffle_indices:
            np.random.shuffle(indices)

        # We add indices to the indices so that workers know where
        # to write in the metadata array
        indices: List[int] = list(enumerate(indices))

        self._write_common(len(indices), chunks(indices, chunksize),
                           worker_job_indexed_dataset, (dataset, ))


    def from_webdataset(self, shards: List[str], pipeline: Callable):
        """Read from webdataset-like format.
        See https://docs.ffcv.io/writing_datasets.html#webdataset for sample usage.

        Parameters
        ----------
        shards: List[str]
            List of shards that comprise the dataset folder.
        pipeline: Callable
            Called by each worker to decode. Similar to pipelines used to load webdataset.
        """
        counter = partial(count_samples_in_shard, pipeline=pipeline)
        lengths = thread_map(counter, shards, max_workers=self.num_workers)
        total_len = sum(lengths)

        offsets = np.cumsum([0] + lengths)[:-1]

        todos = zip(shards, offsets)
        self._write_common(total_len, todos, worker_job_webdataset, (pipeline, ))


    def finalize(self, allocations) :
        # Writing metadata
        with open(self.fname, 'r+b') as fp:
            fp.seek(self.metadata_start)
            fp.write(self.metadata_sm.buf)

            # We go at the end of the file
            fp.seek(0, SEEK_END)

            # Look at the current address
            allocation_table_location = fp.tell()
            # Retrieve all the allocations from the workers
            # Turn them into a numpy array
            try:
                allocation_table = np.array([
                    np.array(x, dtype=ALLOC_TABLE_TYPE) for x in allocations if len(x)
                ])
            except:
                allocation_table = np.array([]).view(ALLOC_TABLE_TYPE)
            fp.write(allocation_table.tobytes())

            self.header['alloc_table_ptr'] = allocation_table_location
            # We go at the start of the file
            fp.seek(0)
            # And write the header
            fp.write(self.header.tobytes())



import numpy as np
import torch as ch
from torch.utils.data import Dataset
from assertpy import assert_that
from tempfile import NamedTemporaryFile
from torchvision.datasets import CIFAR10
from tqdm import tqdm

from ffcv.writer import DatasetWriter
from ffcv.fields import IntField, RGBImageField
from ffcv.fields.decoders import SimpleRGBImageDecoder
from ffcv.loader import Loader
from ffcv.pipeline.compiler import Compiler
from ffcv.transforms import ToTorchImage, ToTensor, NormalizeImage, View, ToDevice

class DummyDataset(Dataset):

    def __init__(self, length, height, width):
        self.length = length
        self.height = height
        self.width = width
    def __len__(self):
        return self.length

    def __getitem__(self, index):
        if index > self.length:
            raise IndexError
        dims = (self.height, self.width, 3)
        image_data = ((np.ones(dims) * index) % 255).astype('uint8')
        result = index,image_data
        return result

def test_cpu_normalization():

    dataset = DummyDataset(500, 25, 30)

    with NamedTemporaryFile() as handle:
        name = handle.name
        fields = {
            'index': IntField(),
            'value': RGBImageField(write_mode='raw', jpeg_quality=95)
        }
        writer = DatasetWriter(name, fields, num_workers=2)
        mean = np.array([0, 1, 2])
        std = np.array([1, 10, 20])

        writer.from_indexed_dataset(dataset, chunksize=5)
        loader = Loader(name, batch_size=5, num_workers=2,
        pipelines={
            'value': [
                SimpleRGBImageDecoder(),
                NormalizeImage(mean, std, np.float16),
                View(np.float16),
                ToTensor(),
                ToTorchImage(),
            ]
        })
        ix = 0
        for res in tqdm(loader):
            index, images  = res
            for image in images:
                image = image.numpy()
                ref_image = dataset[ix][1]
                ref_image = ref_image.transpose(2, 0, 1)
                ref_image = ref_image.astype(np.float16)
                ref_image -= mean[:, None, None]
                ref_image /= std[:, None, None]
                assert_that(np.allclose(ref_image, image)).is_true()
                ix += 1

def test_gpu_normalization():

    dataset = DummyDataset(500, 25, 30)

    with NamedTemporaryFile() as handle:
        name = handle.name
        fields = {
            'index': IntField(),
            'value': RGBImageField(write_mode='raw', jpeg_quality=95)
        }
        writer = DatasetWriter(name, fields, num_workers=2)
        mean = np.array([0, 1, 2])
        std = np.array([1, 10, 20])

        writer.from_indexed_dataset(dataset, chunksize=5)

        loader = Loader(name, batch_size=5, num_workers=2,
        pipelines={
            'value': [
                SimpleRGBImageDecoder(),
                ToTensor(),
                ToDevice(ch.device('cuda:0')),
                ToTorchImage(),
                NormalizeImage(mean, std, np.float16),
                View(ch.float16),
            ]
        })
        ix = 0
        for res in tqdm(loader):
            _, images  = res
            for image in images:
                image = image.cpu().numpy()
                ref_image = dataset[ix][1]
                ref_image = ref_image.transpose(2, 0, 1)
                ref_image = ref_image.astype(np.float16)
                ref_image -= mean[:, None, None]
                ref_image /= std[:, None, None]
                assert_that(np.allclose(ref_image, image)).is_true()
                ix += 1



"""
General operations:

- Collation
- Conversion to PyTorch Tensor
- Change device of Tensor
"""
import torch as ch
import numpy as np
from typing import Callable, Optional, Tuple
from ..pipeline.allocation_query import AllocationQuery
from ..pipeline.operation import Operation
from ..pipeline.state import State
from dataclasses import replace


class ToTensor(Operation):
    """Convert from Numpy array to PyTorch Tensor."""
    def __init__(self):
        super().__init__()

    def generate_code(self) -> Callable:
        def to_tensor(inp, dst):
            return ch.from_numpy(inp)
        return to_tensor

    def declare_state_and_memory(self, previous_state: State) -> Tuple[State, Optional[AllocationQuery]]:
        new_dtype = ch.from_numpy(np.empty((), dtype=previous_state.dtype)).dtype
        return replace(previous_state, jit_mode=False, dtype=new_dtype), None


class ToDevice(Operation):
    """Move tensor to device.

    Parameters
    ----------
    device: torch.device
        Device to move to.
    non_blocking: bool
        Asynchronous if copying from CPU to GPU.
    """
    def __init__(self, device, non_blocking=True):
        super().__init__()
        self.device = device
        # assert isinstance(device, ch.device), \
        #    f'Make sure device is a ch.device (not a {type(device)})'
        self.non_blocking = non_blocking

    def generate_code(self) -> Callable:
        def to_device(inp, dst):
            if len(inp.shape) == 4:
                if inp.is_contiguous(memory_format=ch.channels_last):
                    dst = dst.reshape(inp.shape[0], inp.shape[2], inp.shape[3], inp.shape[1])
                    dst = dst.permute(0, 3, 1, 2)
            dst = dst[:inp.shape[0]]
            dst.copy_(inp, non_blocking=self.non_blocking)
            return dst

        return to_device

    def declare_state_and_memory(self, previous_state: State) -> Tuple[State, Optional[AllocationQuery]]:
        return replace(previous_state, device=self.device), AllocationQuery(previous_state.shape, dtype=previous_state.dtype, device=self.device)


class ToTorchImage(Operation):
    """Change tensor to PyTorch format for images (B x C x H x W).

    Parameters
    ----------
    channels_last : bool
        Use torch.channels_last.
    convert_back_int16 : bool
        Convert to float16.
    """
    def __init__(self, channels_last=True, convert_back_int16=True):
        super().__init__()
        self.channels_last = channels_last
        self.convert_int16 = convert_back_int16
        self.enable_int16conv = False

    def generate_code(self) -> Callable:
        do_conv = self.enable_int16conv
        def to_torch_image(inp: ch.Tensor, dst):
            # Returns a permuted view of the same tensor
            if do_conv:
                inp = inp.view(dtype=ch.float16)
                pass
            inp = inp.permute([0, 3, 1, 2])
            # If channels last, it's already contiguous so we're good
            if self.channels_last:
                assert inp.is_contiguous(memory_format=ch.channels_last)
                return inp

            # Otherwise, need to fill the allocated memory with the contiguous tensor
            dst[:inp.shape[0]] = inp.contiguous()
            return dst[:inp.shape[0]]

        return to_torch_image

    def declare_state_and_memory(self, previous_state: State) -> Tuple[State, Optional[AllocationQuery]]:
        alloc = None
        H, W, C = previous_state.shape
        new_type = previous_state.dtype

        if new_type is ch.int16 and self.convert_int16:
            new_type = ch.float16
            self.enable_int16conv = True

        if not self.channels_last:
            alloc = AllocationQuery((C, H, W), dtype=new_type)
        return replace(previous_state, shape=(C, H, W), dtype=new_type), alloc


class Convert(Operation):
    """Convert to target data type.

    Parameters
    ----------
    target_dtype: numpy.dtype or torch.dtype
        Target data type.
    """
    def __init__(self, target_dtype):
        super().__init__()
        self.target_dtype = target_dtype

    def generate_code(self) -> Callable:
        def convert(inp, dst):
            return inp.type(self.target_dtype)

        convert.is_parallel = True

        return convert

    # TODO: something weird about device to allocate on
    def declare_state_and_memory(self, previous_state: State) -> Tuple[State, Optional[AllocationQuery]]:
        return replace(previous_state, dtype=self.target_dtype), None


class View(Operation):
    """View array using np.view or torch.view.

    Parameters
    ----------
    target_dtype: numpy.dtype or torch.dtype
        Target data type.
    """
    def __init__(self, target_dtype):
        super().__init__()
        self.target_dtype = target_dtype

    def generate_code(self) -> Callable:
        def convert(inp, dst):
            return inp.view(self.target_dtype)

        convert.is_parallel = True

        return convert

    def declare_state_and_memory(self, previous_state: State) -> Tuple[State, Optional[AllocationQuery]]:
        return replace(previous_state, dtype=self.target_dtype, jit_mode=False), None



from typing import Callable, TYPE_CHECKING, Tuple, Type
import warnings
import json
from dataclasses import replace

import numpy as np
import torch as ch

from .base import Field, ARG_TYPE
from ..pipeline.operation import Operation
from ..pipeline.state import State
from ..pipeline.compiler import Compiler
from ..pipeline.allocation_query import AllocationQuery
from ..libffcv import memcpy

if TYPE_CHECKING:
    from ..memory_managers.base import MemoryManager

class NDArrayDecoder(Operation):
    """
    Default decoder for :class:`~ffcv.fields.NDArrayField`.
    """

    def declare_state_and_memory(self, previous_state: State) -> Tuple[State, AllocationQuery]:
        return (
            replace(previous_state, jit_mode=True,
                    shape=self.field.shape,
                    dtype=self.field.dtype),
            AllocationQuery(self.field.shape, self.field.dtype)
        )

    def generate_code(self) -> Callable:
        my_range = Compiler.get_iterator()
        mem_read = self.memory_read
        my_memcpy = Compiler.compile(memcpy)

        def decoder(indices, destination, metadata, storage_state):
            for ix in my_range(indices.shape[0]):
                sample_id = indices[ix]
                ptr = metadata[sample_id]
                data = mem_read(ptr, storage_state)
                my_memcpy(data, destination[ix].view(np.uint8))
            return destination

        return decoder

NDArrayArgsType = np.dtype([
    ('shape', '<u8', 32),  # 32 is the max number of dimensions for numpy
    ('type_length', '<u8'),  # length of the dtype description
])

class NDArrayField(Field):
    """A subclass of :class:`~ffcv.fields.Field` supporting
    multi-dimensional fixed size matrices of any numpy type.
    """
    def __init__(self, dtype:np.dtype, shape:Tuple[int, ...]):
        self.dtype = dtype
        self.shape = shape
        self.element_size = dtype.itemsize * np.prod(shape)
        if dtype == np.uint16:
            warnings.warn("Pytorch currently doesn't support uint16"
            "we recommend storing as int16 and reinterpret your data later"
            "in your pipeline")

    @property
    def metadata_type(self) -> np.dtype:
        return np.dtype('<u8')

    @staticmethod
    def from_binary(binary: ARG_TYPE) -> Field:
        header_size = NDArrayArgsType.itemsize
        header = binary[:header_size].view(NDArrayArgsType)[0]
        type_length = header['type_length']
        type_data = binary[header_size:][:type_length].tobytes().decode('ascii')
        type_desc = json.loads(type_data)
        type_desc = [tuple(x) for x in type_desc]
        assert len(type_desc) == 1
        dtype = np.dtype(type_desc)['f0']
        shape = list(header['shape'])
        while shape[-1] == 0:
            shape.pop()

        return NDArrayField(dtype, tuple(shape))

    def to_binary(self) -> ARG_TYPE:
        result = np.zeros(1, dtype=ARG_TYPE)[0]
        header = np.zeros(1, dtype=NDArrayArgsType)
        s = np.array(self.shape).astype('<u8')
        header['shape'][0][:len(s)] = s
        encoded_type = json.dumps(self.dtype.descr)
        encoded_type = np.frombuffer(encoded_type.encode('ascii'), dtype='<u1')
        header['type_length'][0] = len(encoded_type)
        to_write = np.concatenate([header.view('<u1'), encoded_type])
        result[0][:to_write.shape[0]] = to_write
        return result

    def encode(self, destination, field, malloc):
        destination[0], data_region = malloc(self.element_size)
        data_region[:] = field.reshape(-1).view('<u1')

    def get_decoder_class(self) -> Type[Operation]:
        return NDArrayDecoder


class TorchTensorField(NDArrayField):
    """A subclass of :class:`~ffcv.fields.Field` supporting
    multi-dimensional fixed size matrices of any torch type.
    """
    def __init__(self, dtype:ch.dtype, shape:Tuple[int, ...]):
        self.dtype = dtype
        self.shape = shape
        dtype = ch.zeros(0, dtype=dtype).numpy().dtype

        super().__init__(dtype, shape)


    def encode(self, destination, field, malloc):
        field = field.numpy()
        return super().encode(destination, field, malloc)



"""
FFCV loader
"""
import enum
from os import environ, sched_getaffinity
import ast
from multiprocessing import cpu_count
from re import sub
from typing import Any, Callable, Mapping, Sequence, Type, Union, Literal
from collections import defaultdict
from collections.abc import Collection
from enum import Enum, unique, auto

from ffcv.fields.base import Field

import torch as ch
import numpy as np

from .epoch_iterator import EpochIterator
from ..reader import Reader
from ..traversal_order.base import TraversalOrder
from ..traversal_order import Random, Sequential, QuasiRandom
from ..pipeline import Pipeline, PipelineSpec, Compiler
from ..pipeline.operation import Operation
from ..pipeline.graph import Graph
from ..memory_managers import (
    ProcessCacheManager, OSCacheManager, MemoryManager
)

@unique
class OrderOption(Enum):
    SEQUENTIAL = auto()
    RANDOM = auto()
    QUASI_RANDOM = auto()

ORDER_TYPE = Union[
    TraversalOrder,
    Literal[OrderOption.SEQUENTIAL,
            OrderOption.RANDOM]

]

ORDER_MAP: Mapping[ORDER_TYPE, TraversalOrder] = {
    OrderOption.RANDOM: Random,
    OrderOption.SEQUENTIAL: Sequential,
    OrderOption.QUASI_RANDOM: QuasiRandom
}

DEFAULT_PROCESS_CACHE = int(environ.get('FFCV_DEFAULT_CACHE_PROCESS', "0"))
DEFAULT_OS_CACHE = not DEFAULT_PROCESS_CACHE

class Loader:
    """FFCV loader class that can be used as a drop-in replacement
    for standard (e.g. PyTorch) data loaders.

    Parameters
    ----------
    fname: str
        Full path to the location of the dataset (.beton file format).
    batch_size : int
        Batch size.
    num_workers : int
        Number of workers used for data loading. Consider using the actual number of cores instead of the number of threads if you only use JITed augmentations as they usually don't benefit from hyper-threading.
    os_cache : bool
        Leverages the operating for caching purposes. This is beneficial when there is enough memory to cache the dataset and/or when multiple processes on the same machine training using the same dataset. See https://docs.ffcv.io/performance_guide.html for more information.
    order : Union[OrderOption, TraversalOrder]
        Traversal order, one of: SEQUENTIAL, RANDOM, QUASI_RANDOM, or a custom TraversalOrder

        QUASI_RANDOM is a random order that tries to be as uniform as possible while minimizing the amount of data read from the disk. Note that it is mostly useful when `os_cache=False`. Currently unavailable in distributed mode.
    distributed : bool
        For distributed training (multiple GPUs). Emulates the behavior of DistributedSampler from PyTorch.
    seed : int
        Random seed for batch ordering.
    indices : Sequence[int]
        Subset of dataset by filtering only some indices.
    pipelines : Mapping[str, Sequence[Union[Operation, torch.nn.Module]]
        Dictionary defining for each field the sequence of Decoders and transforms to apply.
        Fileds with missing entries will use the default pipeline, which consists of the default decoder and `ToTensor()`,
        but a field can also be disabled by explicitly by passing `None` as its pipeline.
    custom_fields : Mapping[str, Field]
        Dictonary informing the loader of the types associated to fields that are using a custom type.
    drop_last : bool
        Drop non-full batch in each iteration.
    batches_ahead : int
        Number of batches prepared in advance; balances latency and memory.
    recompile : bool
        Recompile every iteration. This is necessary if the implementation of some augmentations are expected to change during training.
    """
    def __init__(self,
                 fname: str,
                 batch_size: int,
                 num_workers: int = -1,
                 os_cache: bool = DEFAULT_OS_CACHE,
                 order: Union[ORDER_TYPE, TraversalOrder] = OrderOption.SEQUENTIAL,
                 distributed: bool = False,
                 seed: int = None,  # For ordering of samples
                 indices: Sequence[int] = None,  # For subset selection
                 pipelines: Mapping[str,
                                    Sequence[Union[Operation, ch.nn.Module]]] = {},
                 custom_fields: Mapping[str, Type[Field]] = {},
                 drop_last: bool = True,
                 batches_ahead: int = 3,
                 recompile: bool = False,  # Recompile at every epoch
                 order_kwargs: dict = dict(),
                 ):

        if distributed and order == OrderOption.RANDOM and (seed is None):
            print('Warning: no ordering seed was specified with distributed=True. '
                  'Setting seed to 0 to match PyTorch distributed sampler.')
            seed = 0
        elif seed is None:
            tinfo = np.iinfo('int32')
            seed = np.random.randint(0, tinfo.max)

        # We store the original user arguments to be able to pass it to the
        # filtered version of the datasets
        self._args = {
            'fname': fname,
            'batch_size': batch_size,
            'num_workers': num_workers,
            'os_cache': os_cache,
            'order': order,
            'distributed': distributed,
            'seed': seed,
            'indices': indices,
            'pipelines': pipelines,
            'drop_last': drop_last,
            'batches_ahead': batches_ahead,
            'recompile': recompile
        }
        self.fname: str = fname
        self.batch_size: int = batch_size
        self.batches_ahead = batches_ahead
        self.seed: int = seed
        self.reader: Reader = Reader(self.fname, custom_fields)
        self.num_workers: int = num_workers
        self.drop_last: bool = drop_last
        self.distributed: bool = distributed
        self.code = None
        self.recompile = recompile

        if self.num_workers < 1:
            self.num_workers = len(sched_getaffinity(0))

        Compiler.set_num_threads(self.num_workers)

        if indices is None:
            self.indices = np.arange(self.reader.num_samples, dtype='uint64')
        else:
            self.indices = np.array(indices)

        if os_cache:
            self.memory_manager: MemoryManager = OSCacheManager(self.reader)
        else:
            self.memory_manager: MemoryManager = ProcessCacheManager(
                self.reader)

        if order in ORDER_MAP:
            self.traversal_order: TraversalOrder = ORDER_MAP[order](self)
        elif issubclass(order, TraversalOrder):
            self.traversal_order: TraversalOrder = order(self, **order_kwargs)
        else:
            raise ValueError(f"Order {order} is not a supported order type or a subclass of TraversalOrder")

        memory_read = self.memory_manager.compile_reader()
        self.next_epoch: int = 0

        self.pipelines = {}
        self.pipeline_specs = {}
        self.field_name_to_f_ix = {}
        
        custom_pipeline_specs = {}

        # Creating PipelineSpec objects from the pipeline dict passed
        # by the user
        for output_name, spec in pipelines.items():
            if isinstance(spec, PipelineSpec):
                pass
            elif isinstance(spec, Sequence):
                spec = PipelineSpec(output_name, decoder=None, transforms=spec)
            elif spec is None:
                continue  # This is a disabled field
            else:
                msg = f"The pipeline for {output_name} has to be "
                msg += f"either a PipelineSpec or a sequence of operations"
                raise ValueError(msg)
            custom_pipeline_specs[output_name] = spec

        # Adding the default pipelines
        for f_ix, (field_name, field) in enumerate(self.reader.handlers.items()):
            self.field_name_to_f_ix[field_name] = f_ix

            if field_name not in custom_pipeline_specs:
                # We add the default pipeline
                if field_name not in pipelines:
                    self.pipeline_specs[field_name] = PipelineSpec(field_name)
            else:
                self.pipeline_specs[field_name] = custom_pipeline_specs[field_name]

        # We add the custom fields after the default ones
        # This is to preserve backwards compatibility and make sure the order
        # is intuitive
        for field_name, spec in custom_pipeline_specs.items():
            if field_name not in self.pipeline_specs:
                self.pipeline_specs[field_name] = spec

        self.graph = Graph(self.pipeline_specs, self.reader.handlers,
                           self.field_name_to_f_ix, self.reader.metadata,
                           memory_read)
        
        self.generate_code()
        self.first_traversal_order = self.next_traversal_order()

    def next_traversal_order(self):
        return self.traversal_order.sample_order(self.next_epoch)

    def __iter__(self):
        Compiler.set_num_threads(self.num_workers)
        order = self.next_traversal_order()
        selected_order = order[:len(self) * self.batch_size]
        self.next_epoch += 1

        # Compile at the first epoch
        if self.code is None or self.recompile:
            self.generate_code()

        return EpochIterator(self, selected_order)

    def filter(self, field_name:str, condition: Callable[[Any], bool]) -> 'Loader':
        new_args = {**self._args}
        pipelines = {}

        # Disabling all the other fields
        for other_field_name in self.reader.handlers.keys():
            pipelines[other_field_name] = None

        # We reuse the original pipeline for the field we care about
        try:
            pipelines[field_name] = new_args['pipelines'][field_name]
        except KeyError:
            # We keep the default one if the user didn't setup a custom one
            del pipelines[field_name]
            pass

        new_args['pipelines'] = pipelines

        # We use sequential order for speed and to know which index we are
        # filtering
        new_args['order'] = OrderOption.SEQUENTIAL
        new_args['drop_last'] = False
        sub_loader = Loader(**new_args)
        selected_indices = []

        # Iterate through the loader and test the user defined condition
        for i, (batch,) in enumerate(sub_loader):
            for j, sample in enumerate(batch):
                sample_id = i * self.batch_size + j
                if condition(sample):
                    selected_indices.append(sample_id)

        final_args = {**self._args}
        final_args['indices'] = np.array(selected_indices)
        return Loader(**final_args)


    def __len__(self):
        next_order = self.first_traversal_order
        if self.drop_last:
            return len(next_order) // self.batch_size
        else:
            return int(np.ceil(len(next_order) / self.batch_size))



    def generate_code(self):
        queries, code = self.graph.collect_requirements()
        self.code = self.graph.codegen_all(code)
        




from typing import Optional, Sequence, Tuple, Union
from dataclasses import dataclass

import numpy as np
import torch as ch


@dataclass(frozen=True)
class AllocationQuery:
    shape: Tuple[int, ...]
    dtype: Union[np.dtype, ch.dtype]
    device: Optional[ch.device] = None


Allocation = Union[AllocationQuery, Sequence[AllocationQuery]]

def allocate_query(memory_allocation: AllocationQuery, batch_size: int, batches_ahead: int):
    # We compute the total amount of memory needed for this
    # operation
    final_shape = [batches_ahead,
                   batch_size, *memory_allocation.shape]
    if isinstance(memory_allocation.dtype, ch.dtype):
        result = []
        for _ in range(final_shape[0]):
            partial = ch.empty(*final_shape[1:],
                              dtype=memory_allocation.dtype,
                              device=memory_allocation.device)
            try:
                partial = partial.pin_memory()
            except:
                pass
            result.append(partial)
    else:
        ch_dtype = ch.from_numpy(np.empty(0, dtype=memory_allocation.dtype)).dtype
        result = ch.empty(*final_shape,
                          dtype=ch_dtype)
        try:
            result = result.pin_memory()
        except:
            pass
        result = result.numpy()
    return result


import pdb
from numba import njit, set_num_threads, prange, warnings as nwarnings, get_num_threads
from numba.core.errors import NumbaPerformanceWarning
from multiprocessing import cpu_count
import torch as ch
import warnings
from os import sched_getaffinity

class Compiler:

    @classmethod
    def set_enabled(cls, b):
        cls.is_enabled = b

    @classmethod
    def set_num_threads(cls, n):
        if n < 1 :
            n = len(sched_getaffinity(0))
        cls.num_threads = n
        set_num_threads(n)
        ch.set_num_threads(n)

    @classmethod
    def compile(cls, code, signature=None):
        parallel = False
        if hasattr(code, 'is_parallel'):
            parallel = code.is_parallel and cls.num_threads > 1
        
        if cls.is_enabled:
            return njit(signature, fastmath=True, nogil=True, error_model='numpy',
                        parallel=parallel)(code)
        return code

    @classmethod
    def get_iterator(cls):
        if cls.num_threads > 1:
            return prange
        else:
            return range

Compiler.set_enabled(True)
Compiler.set_num_threads(1)



from itertools import product
from time import time
from collections import defaultdict
from contextlib import redirect_stderr
import pathlib

import numpy as np
from tqdm import tqdm

from .benchmark import Benchmark

ALL_SUITES = {}

class FakeSink(object):
    def write(self, *args):
        pass
    def writelines(self, *args):
        pass
    def close(self, *args):
        pass
    def flush(self, *args):
        pass


def benchmark(arg_values={}):
    args_list = product(*arg_values.values())
    runs = [dict(zip(arg_values.keys(), x)) for x in args_list]
    def wrapper(cls):
        ALL_SUITES[cls.__name__] = (cls, runs)
    
    return wrapper

def run_all(runs=3, warm_up=1, pattern='*'):
    results = defaultdict(list)

    selected_suites = {}
    for sname in  ALL_SUITES.keys():
        if pathlib.PurePath(sname).match(pattern):
            selected_suites[sname] = ALL_SUITES[sname]

    it_suite = tqdm(selected_suites.items(), desc='Suite', leave=False)

    for suite_name, (cls, args_list) in it_suite:
        it_suite.set_postfix({'name': suite_name})
        it_args = tqdm(args_list, desc='configuration', leave=False)

        for args in it_args:
            # with redirect_stderr(FakeSink()):
            if True:
                benchmark: Benchmark = cls(**args)
                with benchmark:
                    for _ in range(warm_up):
                        benchmark.run()
                        
                    timings = []
                    for _ in range(runs):
                        start = time()
                        benchmark.run()
                        timings.append(time() - start)
                
            median_time = np.median(timings)
            
            throughput = None
            
            if 'n' in args:
                throughput = args['n'] / median_time
                
            unit = 'it/sec'
            if throughput < 1:
                unit = 'sec/it'
                throughput = 1 /throughput
                
            throughput = np.round(throughput * 10) / 10
             
            results[suite_name].append({
                **args,
                'time': median_time,
                'throughput': str(throughput) + ' ' + unit
            })
        it_args.close()
    it_suite.close()
    return results


"""
Replace label
"""
from typing import Tuple

import numpy as np
from dataclasses import replace
from typing import Callable, Optional, Tuple
from ..pipeline.allocation_query import AllocationQuery
from ..pipeline.operation import Operation
from ..pipeline.state import State
from ..pipeline.compiler import Compiler

class ReplaceLabel(Operation):
    """Replace label of specified images.

    Parameters
    ----------
    indices : Sequence[int]
        The indices of images to relabel.
    new_label : int
        The new label to assign.
    """

    def __init__(self, indices, new_label: int):
        super().__init__()
        self.indices = np.sort(indices)
        self.new_label = new_label

    def generate_code(self) -> Callable:

        to_change = self.indices
        new_label = self.new_label
        my_range = Compiler.get_iterator()

        def replace_label(labels, temp_array, indices):
            for i in my_range(labels.shape[0]):
                sample_ix = indices[i]
                position = np.searchsorted(to_change, sample_ix)
                if position < len(to_change) and to_change[position] == sample_ix:
                    labels[i] = new_label
            return labels

        replace_label.is_parallel = True
        replace_label.with_indices = True

        return replace_label

    def declare_state_and_memory(self, previous_state: State) -> Tuple[State, Optional[AllocationQuery]]:
        return (replace(previous_state, jit_mode=True), None)




import numpy as np
from tqdm import tqdm
from assertpy import assert_that
from torch.utils.data import Dataset
import logging
from time import time
import os
from assertpy import assert_that
from tempfile import NamedTemporaryFile

from ffcv.writer import DatasetWriter
from ffcv.reader import Reader
from ffcv.fields import BytesField, IntField
from ffcv.pipeline.compiler import Compiler
from ffcv.memory_managers import OSCacheManager

from test_memory_allocation import DummyDataset


def create_and_validate(length, size, do_compile):

    dataset = DummyDataset(length, size)

    with NamedTemporaryFile() as handle:
        name = handle.name
        writer = DatasetWriter(name, {
            'index': IntField(),
            'value': BytesField()
        }, num_workers=2)

        writer.from_indexed_dataset(dataset, chunksize=5)

        reader = Reader(name)
        manager = OSCacheManager(reader)
        context = manager.schedule_epoch(np.array([0, 1]))

        indices = np.random.choice(length, 500)
        addresses = reader.alloc_table['ptr'][indices]
        sample_ids = reader.alloc_table['sample_id'][indices]

        Compiler.set_enabled(do_compile)
        read_fn = manager.compile_reader()

        with context:

            for addr, sample_id in zip(tqdm(addresses), sample_ids):
                read_buffer = read_fn(addr, context.state)
                np.random.seed(sample_id)
                expected_buff = np.random.randint(0, 255, size=size, dtype='u1')

                assert_that(read_buffer).is_length(len(expected_buff))
                assert_that(np.all(read_buffer == expected_buff)).is_true()

            # We skip the first which is compilation

def test_simple():
    create_and_validate(600, 76, False)

def test_large():
    create_and_validate(600, 1024, False)

def test_many():
    create_and_validate(60000, 81, False)

def test_many_compiled():
    create_and_validate(1000000, 1, True)



import random
from typing import Sequence, TYPE_CHECKING
from numba import njit
import numpy as np

from torch.utils.data import DistributedSampler

from .base import TraversalOrder

if TYPE_CHECKING:
    from ..loader.loader import Loader


@njit(parallel=False)
def generate_order_inner(seed, page_to_samples_array, page_sizes,
                         result, buffer_size=6):
    num_pages = len(page_sizes)
    random.seed(seed)
    np.random.seed(seed)
    current_pages = [0]
    current_pages.remove(0)  # Force the type
    for page_ix in range(num_pages):
        page_size = page_sizes[page_ix]
        random.shuffle(page_to_samples_array[page_ix, :page_size])
    next_page = 0
    page_order = np.random.permutation(num_pages)
    samples_consumed = np.zeros_like(page_sizes)
    for s_ix in range(result.shape[0]):
        while next_page < num_pages and len(current_pages) < buffer_size:
            page_to_add = page_order[next_page]
            if page_sizes[page_to_add] > 0:
                current_pages.append(page_order[next_page])
            next_page += 1
        selected_page_ix = np.random.randint(0, len(current_pages))
        page = current_pages[selected_page_ix]
        result[s_ix] = page_to_samples_array[page, samples_consumed[page]]
        samples_consumed[page] += 1
        if samples_consumed[page] >= page_sizes[page]:
            current_pages.remove(page)


class QuasiRandom(TraversalOrder):

    def __init__(self, loader: 'Loader'):
        super().__init__(loader)

        # TODO filter only the samples we care about!!
        self.page_to_samples = loader.memory_manager.page_to_samples

        if not self.page_to_samples:
            raise ValueError(
                "Dataset won't benefit from QuasiRandom order, use regular Random")

        if self.distributed:
            raise NotImplementedError(
                "distributed Not implemented yet for QuasiRandom")

        self.prepare_data_structures()


    def prepare_data_structures(self):
        index_set = set(self.indices)
        max_size = max(len(y) for y in self.page_to_samples.values())
        num_pages = max(k for k in self.page_to_samples.keys()) + np.uint64(1)

        self.page_to_samples_array = np.empty((num_pages, max_size),
                                              dtype=np.int64)
        self.page_sizes = np.zeros(num_pages, dtype=np.int64)

        for page, content in self.page_to_samples.items():
            for c in content:
                if c in index_set:
                    self.page_to_samples_array[page][self.page_sizes[page]] = c
                    self.page_sizes[page] += 1



    def sample_order(self, epoch: int) -> Sequence[int]:
        seed = self.seed * 912300 + epoch
        result_order = np.zeros(len(self.indices), dtype=np.int64)
        generate_order_inner(seed, self.page_to_samples_array,
                             self.page_sizes,
                             result_order,
                             2*self.loader.batch_size)

        return result_order


from typing import Callable, TYPE_CHECKING, Tuple, Type
from dataclasses import replace

import numpy as np

from .base import Field, ARG_TYPE
from ..pipeline.operation import Operation
from ..pipeline.state import State
from ..pipeline.allocation_query import AllocationQuery

if TYPE_CHECKING:
    from ..memory_managers.base import MemoryManager

class BasicDecoder(Operation):
    """For decoding scalar fields

    This Decoder can be extend to decode any fixed length numpy data type
    """
    def declare_state_and_memory(self, previous_state: State) -> Tuple[State, AllocationQuery]:
        my_shape = (1,)
        return (
            replace(previous_state, jit_mode=True,
                    shape=my_shape,
                    dtype=self.dtype),
            AllocationQuery(my_shape, dtype=self.dtype)
        )

    def generate_code(self) -> Callable:
        def decoder(indices, destination, metadata, storage_state):
            for ix, sample_id in enumerate(indices):
                destination[ix] = metadata[sample_id]
            return destination[:len(indices)]

        return decoder

class IntDecoder(BasicDecoder):
    """Decoder for signed integers scalars (int64)
    """
    dtype = np.dtype('<i8')

class FloatDecoder(BasicDecoder):
    """Decoder for floating point scalars (float64)
    """
    dtype = np.dtype('<f8')

class FloatField(Field):
    """
    A subclass of :class:`~ffcv.fields.Field` supporting (scalar) floating-point (float64)
    values.
    """
    def __init__(self):
        pass

    @property
    def metadata_type(self) -> np.dtype:
        return np.dtype('<f8')

    @staticmethod
    def from_binary(binary: ARG_TYPE) -> Field:
        return FloatField()

    def to_binary(self) -> ARG_TYPE:
        return np.zeros(1, dtype=ARG_TYPE)[0]

    def encode(self, destination, field, malloc):
        destination[0] = field

    def get_decoder_class(self) -> Type[Operation]:
        return FloatDecoder

class IntField(Field):
    """
    A subclass of :class:`~ffcv.fields.Field` supporting (scalar) integer
    values.
    """
    @property
    def metadata_type(self) -> np.dtype:
        return np.dtype('<i8')

    @staticmethod
    def from_binary(binary: ARG_TYPE) -> Field:
        return IntField()

    def to_binary(self) -> ARG_TYPE:
        return np.zeros(1, dtype=ARG_TYPE)[0]

    def encode(self, destination, field, malloc):
        # We just allocate 1024bytes for fun
        destination[0] = field

    def get_decoder_class(self) -> Type[Operation]:
        return IntDecoder




from typing import Sequence, TYPE_CHECKING
import numpy as np

from torch.utils.data import DistributedSampler

from .base import TraversalOrder

if TYPE_CHECKING:
    from ..loader.loader import Loader
    

class Sequential(TraversalOrder):
    
    def __init__(self, loader:'Loader'):
        super().__init__(loader)
        
        if self.distributed:
            self.sampler = DistributedSampler(self.indices,
                                              shuffle=False,
                                              seed=self.seed,
                                              drop_last=False)
        

    def sample_order(self, epoch: int) -> Sequence[int]:
        if not self.distributed:
            return self.indices
        
        self.sampler.set_epoch(epoch)
        
        return self.indices[np.array(list(self.sampler))]



"""
Example of using FFCV to speed up large scale linear regression.
For tutorial, see https://docs.ffcv.io/ffcv_examples/linear_regression.html.

"""
from tqdm import tqdm
import time
import numpy as np
import pickle as pkl
import torch as ch
from torch.utils.data import TensorDataset, DataLoader
from ffcv.fields import NDArrayField, FloatField
from ffcv.fields.basics import FloatDecoder
from ffcv.fields.decoders import NDArrayDecoder
from ffcv.loader import Loader, OrderOption
from ffcv.writer import DatasetWriter
from ffcv.transforms import ToTensor, ToDevice, Squeeze
import os

# 1,000,000 inputs each of dimension 10,000 = 40GB of data
N, D = 1000000, 10000
USE_FFCV = True
if not os.path.exists('/tmp/linreg_data.pkl'):
    X = np.random.rand(N, D).astype('float32')
    # Ground-truth vector
    W, b = np.random.rand(D).astype('float32'), np.random.rand()
    # Response variables
    Y = X @ W + b + np.random.randn(N).astype('float32')
    pkl.dump((X, W, b, Y), open('/tmp/linreg_data.pkl', 'wb'))
elif not USE_FFCV:
    print('Loading from disk...')
    X, W, b, Y = pkl.load(open('/tmp/linreg_data.pkl', 'rb'))

if USE_FFCV and not os.path.exists('/tmp/linreg_data.beton'):
    X, W, b, Y = pkl.load(open('/tmp/linreg_data.pkl', 'rb'))
    class LinearRegressionDataset:
        def __getitem__(self, idx):
            return (X[idx], np.array(Y[idx]).astype('float32'))

        def __len__(self):
            return len(X)

    writer = DatasetWriter('/tmp/linreg_data.beton', {
        'covariate': NDArrayField(shape=(D,), dtype=np.dtype('float32')),
        'label': NDArrayField(shape=(1,), dtype=np.dtype('float32')),
    }, num_workers=16)

    writer.from_indexed_dataset(LinearRegressionDataset())
else:
    print('FFCV file already written')


### PART 2: actual regression

if not USE_FFCV:
    dataset = TensorDataset(ch.tensor(X), ch.tensor(Y))
    train_loader = DataLoader(dataset, batch_size=2048, num_workers=8, shuffle=True)
else:
    train_loader = Loader('/tmp/linreg_data.beton', batch_size=2048,
                    num_workers=8, order=OrderOption.QUASI_RANDOM, os_cache=False,
                    pipelines={
                        'covariate': [NDArrayDecoder(), ToTensor(), ToDevice(ch.device('cuda:0'))],
                        'label': [NDArrayDecoder(), ToTensor(), Squeeze(), ToDevice(ch.device('cuda:0'))]
                    })

# Calculate data mean and variance for normalization
def calculate_stats(loader, N):
    mean, stdev = 0., 0.
    for x_batch, _ in tqdm(loader):
        mean += x_batch.sum(0) / N
        stdev += x_batch.pow(2).sum(0) / N
    return mean, ch.sqrt(stdev - mean.pow(2))

mean, stdev = calculate_stats(train_loader, N)
mean, stdev = mean.cuda(), stdev.cuda()
w_est, b_est = ch.zeros(D).cuda(), ch.zeros(1).cuda() # Initial guess for W
num_epochs = 10 # Number of full passes over the data to do

lr = 5e-2
for _ in range(num_epochs):
    total_loss, num_examples = 0., 0.
    start_time = time.time()
    for (x_batch, y_batch) in tqdm(train_loader):
        if not USE_FFCV:
            x_batch = x_batch.cuda()
            y_batch = y_batch.cuda()
        # Normalize the data for stability
        x_batch = (x_batch - mean) / stdev
        residual = x_batch @ w_est + b_est - y_batch
        # Gradients
        w_grad = x_batch.T @ residual / x_batch.shape[0]
        b_grad = ch.mean(residual, dim=0)
        w_est = w_est - lr * w_grad
        b_est = b_est - lr * b_grad
        total_loss += residual.pow(2).sum()
        num_examples += x_batch.shape[0]
    print('Epoch time:', time.time() - start_time)
    print(f'Average loss: {total_loss / num_examples:.3f}')



from abc import ABC, abstractmethod
from typing import Sequence
from ..reader import Reader

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..loader.main_thread import Loader

class TraversalOrder(ABC):
    def __init__(self, loader: 'Loader'):
        self.loader = loader
        self.indices = self.loader.indices
        self.seed = self.loader.seed
        self.distributed = loader.distributed
        self.sampler = None

    @abstractmethod
    def sample_order(self, epoch:int) -> Sequence[int]:
        raise NotImplemented()



from abc import ABCMeta, abstractmethod
from dataclasses import replace
from typing import Optional, Callable, TYPE_CHECKING, Tuple, Type

import cv2
import numpy as np
from numba.typed import Dict
from PIL.Image import Image

from .base import Field, ARG_TYPE
from ..pipeline.operation import Operation
from ..pipeline.state import State
from ..pipeline.compiler import Compiler
from ..pipeline.allocation_query import AllocationQuery
from ..libffcv import imdecode, memcpy, resize_crop

if TYPE_CHECKING:
    from ..memory_managers.base import MemoryManager
    from ..reader import Reader

IMAGE_MODES = Dict()
IMAGE_MODES['jpg'] = 0
IMAGE_MODES['raw'] = 1


def encode_jpeg(numpy_image, quality):
    numpy_image = cv2.cvtColor(numpy_image, cv2.COLOR_RGB2BGR)
    success, result = cv2.imencode('.jpg', numpy_image,
                                   [int(cv2.IMWRITE_JPEG_QUALITY), quality])

    if not success:
        raise ValueError("Impossible to encode image in jpeg")

    return result.reshape(-1)


def resizer(image, target_resolution):
    if target_resolution is None:
        return image
    original_size = np.array([image.shape[1], image.shape[0]])
    ratio = target_resolution / original_size.max()
    if ratio < 1:
        new_size = (ratio * original_size).astype(int)
        image = cv2.resize(image, tuple(new_size), interpolation=cv2.INTER_AREA)
    return image


def get_random_crop(height, width, scale, ratio):
    area = height * width
    log_ratio = np.log(ratio)
    for _ in range(10):
        target_area = area * np.random.uniform(scale[0], scale[1])
        aspect_ratio = np.exp(np.random.uniform(log_ratio[0], log_ratio[1]))
        w = int(round(np.sqrt(target_area * aspect_ratio)))
        h = int(round(np.sqrt(target_area / aspect_ratio)))
        if 0 < w <= width and 0 < h <= height:
            i = int(np.random.uniform(0, height - h + 1))
            j = int(np.random.uniform(0, width - w + 1))
            return i, j, h, w
    in_ratio = float(width) / float(height)
    if in_ratio < min(ratio):
        w = width
        h = int(round(w / min(ratio)))
    elif in_ratio > max(ratio):
        h = height
        w = int(round(h * max(ratio)))
    else:
        w = width
        h = height
    i = (height - h) // 2
    j = (width - w) // 2
    return i, j, h, w


def get_center_crop(height, width, _, ratio):
    s = min(height, width)
    c = int(ratio * s)
    delta_h = (height - c) // 2
    delta_w = (width - c) // 2

    return delta_h, delta_w, c, c


class SimpleRGBImageDecoder(Operation):
    """Most basic decoder for the :class:`~ffcv.fields.RGBImageField`.

    It only supports dataset with constant image resolution and will simply read (potentially decompress) and pass the images as is.
    """
    def __init__(self):
        super().__init__()

    def declare_state_and_memory(self, previous_state: State) -> Tuple[State, AllocationQuery]:
        widths = self.metadata['width']
        heights = self.metadata['height']
        max_width = widths.max()
        max_height = heights.max()
        min_height = heights.min()
        min_width = widths.min()
        if min_width != max_width or max_height != min_height:
            msg = """SimpleRGBImageDecoder only supports constant image,
consider RandomResizedCropRGBImageDecoder or CenterCropRGBImageDecoder
instead."""
            raise TypeError(msg)

        biggest_shape = (max_height, max_width, 3)
        my_dtype = np.dtype('<u1')

        return (
            replace(previous_state, jit_mode=True,
                    shape=biggest_shape, dtype=my_dtype),
            AllocationQuery(biggest_shape, my_dtype)
        )

    def generate_code(self) -> Callable:
        mem_read = self.memory_read
        imdecode_c = Compiler.compile(imdecode)

        jpg = IMAGE_MODES['jpg']
        raw = IMAGE_MODES['raw']
        my_range = Compiler.get_iterator()
        my_memcpy = Compiler.compile(memcpy)

        def decode(batch_indices, destination, metadata, storage_state):
            for dst_ix in my_range(len(batch_indices)):
                source_ix = batch_indices[dst_ix]
                field = metadata[source_ix]
                image_data = mem_read(field['data_ptr'], storage_state)
                height, width = field['height'], field['width']

                if field['mode'] == jpg:
                    imdecode_c(image_data, destination[dst_ix],
                               height, width, height, width, 0, 0, 1, 1, False, False)
                else:
                    my_memcpy(image_data, destination[dst_ix])

            return destination[:len(batch_indices)]

        decode.is_parallel = True
        return decode


class ResizedCropRGBImageDecoder(SimpleRGBImageDecoder, metaclass=ABCMeta):
    """Abstract decoder for :class:`~ffcv.fields.RGBImageField` that performs a crop and and a resize operation.

    It supports both variable and constant resolution datasets.
    """
    def __init__(self, output_size):
        super().__init__()
        self.output_size = output_size

    def declare_state_and_memory(self, previous_state: State) -> Tuple[State, AllocationQuery]:
        widths = self.metadata['width']
        heights = self.metadata['height']
        # We convert to uint64 to avoid overflows
        self.max_width = np.uint64(widths.max())
        self.max_height = np.uint64(heights.max())
        output_shape = (self.output_size[0], self.output_size[1], 3)
        my_dtype = np.dtype('<u1')

        return (
            replace(previous_state, jit_mode=True,
                    shape=output_shape, dtype=my_dtype),
            (AllocationQuery(output_shape, my_dtype),
            AllocationQuery((self.max_height * self.max_width * np.uint64(3),), my_dtype),
            )
        )

    def generate_code(self) -> Callable:

        jpg = IMAGE_MODES['jpg']

        mem_read = self.memory_read
        my_range = Compiler.get_iterator()
        imdecode_c = Compiler.compile(imdecode)
        resize_crop_c = Compiler.compile(resize_crop)
        get_crop_c = Compiler.compile(self.get_crop_generator)

        scale = self.scale
        ratio = self.ratio
        if isinstance(scale, tuple):
            scale = np.array(scale)
        if isinstance(ratio, tuple):
            ratio = np.array(ratio)

        def decode(batch_indices, my_storage, metadata, storage_state):
            destination, temp_storage = my_storage
            for dst_ix in my_range(len(batch_indices)):
                source_ix = batch_indices[dst_ix]
                field = metadata[source_ix]
                image_data = mem_read(field['data_ptr'], storage_state)
                height = np.uint32(field['height'])
                width = np.uint32(field['width'])

                if field['mode'] == jpg:
                    temp_buffer = temp_storage[dst_ix]
                    imdecode_c(image_data, temp_buffer,
                               height, width, height, width, 0, 0, 1, 1, False, False)
                    selected_size = 3 * height * width
                    temp_buffer = temp_buffer.reshape(-1)[:selected_size]
                    temp_buffer = temp_buffer.reshape(height, width, 3)

                else:
                    temp_buffer = image_data.reshape(height, width, 3)

                i, j, h, w = get_crop_c(height, width, scale, ratio)

                resize_crop_c(temp_buffer, i, i + h, j, j + w,
                              destination[dst_ix])

            return destination[:len(batch_indices)]
        decode.is_parallel = True
        return decode

    @property
    @abstractmethod
    def get_crop_generator():
        raise NotImplementedError


class RandomResizedCropRGBImageDecoder(ResizedCropRGBImageDecoder):
    """Decoder for :class:`~ffcv.fields.RGBImageField` that performs a Random crop and and a resize operation.

    It supports both variable and constant resolution datasets.

    Parameters
    ----------
    output_size : Tuple[int]
        The desired resized resolution of the images
    scale : Tuple[float]
        The range of possible ratios (in area) than can randomly sampled
    ratio : Tuple[float]
        The range of potential aspect ratios that can be randomly sampled
    """
    def __init__(self, output_size, scale=(0.08, 1.0), ratio=(0.75, 4/3)):
        super().__init__(output_size)
        self.scale = scale
        self.ratio = ratio
        self.output_size = output_size

    @property
    def get_crop_generator(self):
        return get_random_crop


class CenterCropRGBImageDecoder(ResizedCropRGBImageDecoder):
    """Decoder for :class:`~ffcv.fields.RGBImageField` that performs a center crop followed by a resize operation.

    It supports both variable and constant resolution datasets.

    Parameters
    ----------
    output_size : Tuple[int]
        The desired resized resolution of the images
    ratio: float
        ratio of (crop size) / (min side length)
    """
    # output size: resize crop size -> output size
    def __init__(self, output_size, ratio):
        super().__init__(output_size)
        self.scale = None
        self.ratio = ratio

    @property
    def get_crop_generator(self):
        return get_center_crop


class RGBImageField(Field):
    """
    A subclass of :class:`~ffcv.fields.Field` supporting RGB image data.

    Parameters
    ----------
    write_mode : str, optional
        How to write the image data to the dataset file. Should be either 'raw'
        (``uint8`` pixel values), 'jpg' (compress to JPEG format), 'smart'
        (decide between saving pixel values and JPEG compressing based on image
        size), and 'proportion' (JPEG compress a random subset of the data with
        size specified by the ``compress_probability`` argument). By default: 'raw'.
    max_resolution : int, optional
        If specified, will resize images to have maximum side length equal to
        this value before saving, by default None
    smart_threshold : int, optional
        When `write_mode='smart`, will compress an image if it would take more than `smart_threshold` times to use RAW instead of jpeg.
    jpeg_quality : int, optional
        The quality parameter for JPEG encoding (ignored for
        ``write_mode='raw'``), by default 90
    compress_probability : float, optional
        Ignored unless ``write_mode='proportion'``; in the latter case it is the
        probability with which image is JPEG-compressed, by default 0.5.
    """
    def __init__(self, write_mode='raw', max_resolution: int = None,
                 smart_threshold: int = None, jpeg_quality: int = 90,
                 compress_probability: float = 0.5) -> None:
        self.write_mode = write_mode
        self.smart_threshold = smart_threshold
        self.max_resolution = max_resolution
        self.jpeg_quality = int(jpeg_quality)
        self.proportion = compress_probability

    @property
    def metadata_type(self) -> np.dtype:
        return np.dtype([
            ('mode', '<u1'),
            ('width', '<u2'),
            ('height', '<u2'),
            ('data_ptr', '<u8'),
        ])

    def get_decoder_class(self) -> Type[Operation]:
        return SimpleRGBImageDecoder

    @staticmethod
    def from_binary(binary: ARG_TYPE) -> Field:
        return RGBImageField()

    def to_binary(self) -> ARG_TYPE:
        return np.zeros(1, dtype=ARG_TYPE)[0]

    def encode(self, destination, image, malloc):
        if isinstance(image, Image):
            image = np.array(image)

        if not isinstance(image, np.ndarray):
            raise TypeError(f"Unsupported image type {type(image)}")

        if image.dtype != np.uint8:
            raise ValueError("Image type has to be uint8")

        if image.shape[2] != 3:
            raise ValueError(f"Invalid shape for rgb image: {image.shape}")

        assert image.dtype == np.uint8

        image = resizer(image, self.max_resolution)

        write_mode = self.write_mode
        as_jpg = None

        if write_mode == 'smart':
            as_jpg = encode_jpeg(image, self.jpeg_quality)
            write_mode = 'raw'
            if self.smart_threshold is not None:
                if image.nbytes > self.smart_threshold:
                    write_mode = 'jpg'
        elif write_mode == 'proportion':
            if np.random.rand() < self.proportion:
                write_mode = 'jpg'
            else:
                write_mode = 'raw'

        destination['mode'] = IMAGE_MODES[write_mode]
        destination['height'], destination['width'] = image.shape[:2]

        if write_mode == 'jpg':
            if as_jpg is None:
                as_jpg = encode_jpeg(image, self.jpeg_quality)
            destination['data_ptr'], storage = malloc(as_jpg.nbytes)
            storage[:] = as_jpg
        elif write_mode == 'raw':
            image_bytes = np.ascontiguousarray(image).view('<u1').reshape(-1)
            destination['data_ptr'], storage = malloc(image.nbytes)
            storage[:] = image_bytes
        else:
            raise ValueError(f"Unsupported write mode {self.write_mode}")



from ffcv.transforms.ops import ToTensor
from ffcv.fields.rgb_image import RandomResizedCropRGBImageDecoder, SimpleRGBImageDecoder, CenterCropRGBImageDecoder
import numpy as np
import torch as ch
from torch.utils.data import Dataset
from assertpy import assert_that
from tempfile import NamedTemporaryFile
from torchvision.datasets import CIFAR10
from torch.utils.data import Subset

from ffcv.writer import DatasetWriter
from ffcv.fields import IntField, RGBImageField
from ffcv.loader import Loader
from ffcv.pipeline.compiler import Compiler

class DummyDataset(Dataset):

    def __init__(self, length, size_range):
        self.length = length
        self.size_range = size_range
        
    def __len__(self):
        return self.length

    def __getitem__(self, index):
        if index > self.length:
            raise IndexError
        dims = (
            np.random.randint(self.size_range[0], self.size_range[1] + 1),
            np.random.randint(self.size_range[0], self.size_range[1] + 1),
            3
        )
        image_data = ((np.ones(dims) * index) % 255).astype('uint8')
        return index, image_data

def create_and_validate(length, decoder, size, mode='raw', compile=False):

    dataset = DummyDataset(length, (300, 500))

    with NamedTemporaryFile() as handle:
        name = handle.name
        
        fields = {
            'index': IntField(),
            'value': RGBImageField(write_mode=mode)
        }

        writer = DatasetWriter(name, fields, num_workers=2)

        writer.from_indexed_dataset(dataset, chunksize=5)
            
        Compiler.set_enabled(compile)
        
        loader = Loader(name, batch_size=5, num_workers=2,
                        pipelines={
                            'value': [decoder, ToTensor()]
                        })
        
        for index, images in loader:
            for i, image in zip(index, images):
                assert_that(image.shape).is_equal_to((size[0], size[1], 3))
                if mode == 'raw':
                    assert_that(ch.all((image == (i % 255)).reshape(-1))).is_true()
                else:
                    assert_that(ch.all(ch.abs(image - (i % 255)) < 2)).is_true
                

def test_simple_image_decoder_fails_with_variable_images():
    decoder = SimpleRGBImageDecoder()
    assert_that(create_and_validate).raises(TypeError).when_called_with(500, decoder, 32, 'raw')

def test_rrc_decoder_raw():
    size = (160, 160)
    decoder = RandomResizedCropRGBImageDecoder(size)
    create_and_validate(500, decoder, size, 'raw')

def test_rrc_decoder_jpg():
    size = (160, 160)
    decoder = RandomResizedCropRGBImageDecoder(size)
    create_and_validate(500, decoder, size, 'jpg')

def test_rrc_decoder_raw_compiled():
    size = (160, 160)
    decoder = RandomResizedCropRGBImageDecoder(size)
    create_and_validate(500, decoder, size, 'raw', True)

def test_rrc_decoder_jpg_compiled():
    size = (160, 160)
    decoder = RandomResizedCropRGBImageDecoder(size)
    create_and_validate(500, decoder, size, 'jpg', True)

def test_cc_decoder_raw_nc():
    size = (160, 160)
    decoder = CenterCropRGBImageDecoder(size, 224/256)
    create_and_validate(500, decoder, size, 'raw')

def test_cc_decoder_jpg_nc():
    size = (160, 160)
    decoder = CenterCropRGBImageDecoder(size, 224/256)
    create_and_validate(500, decoder, size, 'jpg')

def test_cc_decoder_raw_compiled():
    size = (160, 160)
    decoder = CenterCropRGBImageDecoder(size, 224/256)
    create_and_validate(500, decoder, size, 'raw', True)

def test_cc_decoder_jpg_compiled():
    size = (160, 160)
    decoder = CenterCropRGBImageDecoder(size, 224/256)
    create_and_validate(500, decoder, size, 'jpg', True)


if __name__ == '__main__':
    test_rrc_decoder_jpg()


from .base import Field
from .basics import FloatField, IntField
from .rgb_image import RGBImageField
from .bytes import BytesField
from .ndarray import NDArrayField, TorchTensorField
from .json import JSONField

__all__ = ['Field', 'BytesField', 'IntField', 'FloatField', 'RGBImageField',
           'NDArrayField', 'JSONField', 'TorchTensorField']


import json

import torch as ch
import numpy as np

from .bytes import BytesField

ENCODING = 'utf8'
SEPARATOR = '\0'  # Null byte

class JSONField(BytesField):
    """A subclass of :class:`~ffcv.fields.BytesField` that encodes JSON data.

    The writer expects to be passed a dict that is compatible with the JSON specification.

    .. warning ::
        Because FFCV is based on tensors/ndarrays the reader and therefore the loader can't give return JSON to the user. This is why we provide :class:`~ffcv.fields.JSONField.unpack` which does the conversion. It's up to the user to call it in the main body of the loop

    """

    @property
    def metadata_type(self) -> np.dtype:
        return np.dtype([
            ('ptr', '<u8'),
            ('size', '<u8')
        ])

    def encode(self, destination, field, malloc):
        # Add null terminating byte
        content = (json.dumps(field) + SEPARATOR).encode(ENCODING)
        field = np.frombuffer(content, dtype='uint8')
        return super().encode(destination, field, malloc)

    @staticmethod
    def unpack(batch):
        """Convert back the output of a :class:`~ffcv.fields.JSONField` field produced by :class:`~ffcv.Loader` into an actual JSON.
        
        It works both on an entire batch and will return an array of python dicts or a single sample and will simply return a dict.
        """
        if isinstance(batch, ch.Tensor):
            batch = batch.numpy()

        single_instance = len(batch.shape) == 1
        if single_instance:
            batch = [batch]

        result = []
        for b in batch:
            sep_location = np.where(b == ord(SEPARATOR))[0][0]
            b = b[:sep_location]
            string = b.tobytes().decode(ENCODING)
            result.append(json.loads(string))

        if single_instance:
            result = result[0]

        return result



import numba as nb
import numpy as np

from .context import ProcessCacheContext
from ...pipeline.compiler import Compiler
from ..base import MemoryManager, MemoryContext
from ..common import BATCHES_TYPE

class ProcessCacheManager(MemoryManager):

    def schedule_epoch(self, batches: BATCHES_TYPE) -> MemoryContext:
        return ProcessCacheContext(self, batches)

    @property
    def state_type(self):
        # The data
        t1 = nb.uint8[:, ::1]
        t1.mutable = False

        # The pointers
        t2 = nb.uint64[::1]
        t2.mutable = False
        #
        # Their size
        t3 = nb.uint64[::1]
        t3.mutable = False

        # Page to slot
        t4 = nb.uint32[::1]
        t4.mutable = False

        return nb.types.Tuple([t1, t2, t3, t4])

    def compile_reader(self):
        page_size = self.reader.page_size
        page_size_log2 = np.uint32(np.log2(page_size))

        def read(address, mem_state):
            size = mem_state[2][np.searchsorted(mem_state[1], address)]
            page = address >> page_size_log2
            offset = address - (page << page_size_log2)
            page_slot = mem_state[3][page]
            return mem_state[0][page_slot, offset:offset + size]

        return Compiler.compile(read)



import numpy as np

import pytest
from assertpy import assert_that

from ffcv.libffcv import memcpy


MEMCPY_TYPES = [
    np.uint8,
    np.int32,
    np.int64,
    np.float64,
    np.float32
]

@pytest.mark.parametrize('dtype', MEMCPY_TYPES)
def test_memcpy(dtype):

    data = np.random.uniform(0, 255, size=(100, 99)).astype(dtype)
    dst = np.empty((100, 99), dtype=dtype)

    assert_that(np.all(data == dst)).is_false()
    memcpy(data, dst)

    assert_that(np.all(data == dst)).is_true()


from collections import defaultdict
from tempfile import TemporaryDirectory
from os import path
from typing import Counter

import pytest
from assertpy import assert_that
import numpy as np
from torch.utils.data import Dataset, distributed
from torch.multiprocessing import spawn, Queue
from torch.distributed import init_process_group

from ffcv.loader.loader import ORDER_TYPE, OrderOption
from ffcv.writer import DatasetWriter
from ffcv.fields import IntField, BytesField
from ffcv import Loader

class DummyDataset(Dataset):

    def __init__(self, l):
        self.l = l

    def __len__(self):
        return self.l

    def __getitem__(self, index):
        if index > self.l:
            raise IndexError()
        return (index, np.sin(np.array([index])).view('<u1'))


def process_work(rank, world_size, fname, order, sync_fname, out_folder, indices):
    sync_url = f'file://{sync_fname}'
    if world_size > 1:
        init_process_group('nccl', sync_url, rank=rank, world_size=world_size)
    
    loader = Loader(fname, 8, num_workers=2, order=order, drop_last=False,
                    distributed=world_size > 1, indices=indices)
    
    result = []
    for _ in range(3):
        content = np.concatenate([x[0].numpy().reshape(-1).copy() for x in loader])
        result.append(content)
    result = np.stack(result)
    
    np.save(path.join(out_folder, f"result-{rank}.npy"), result)
    

def prep_and_run_test(num_workers, order, with_indices=False):
    length = 600
    indices = None
    if with_indices:
        indices = np.random.choice(length, length//2, replace=False)

    with TemporaryDirectory() as folder:
            name = path.join(folder, 'dataset.beton')
            sync_file = path.join(folder, 'share')
            dataset = DummyDataset(length)
            writer = DatasetWriter(name, {
                'index': IntField(),
                'value': BytesField()
            })

            writer.from_indexed_dataset(dataset)
                
            args = (num_workers, name, order, sync_file, folder, indices)
            if num_workers > 1:
                spawn(process_work, nprocs=num_workers, args=args)
            else:
                process_work(*((0, ) + args))
            
            results = []
            for r in range(num_workers):
                array = np.load(path.join(folder,f"result-{r}.npy"))
                results.append(array)

            results = np.concatenate(results, 1)
            
            # For each epoch
            for i in range(results.shape[0]):
                if not with_indices:
                    if order == OrderOption.SEQUENTIAL and i < results.shape[0] - 1:
                        assert_that((results[i] == results[i + 1]).all()).is_true()
                    if order != OrderOption.SEQUENTIAL and i < results.shape[0] - 1:
                        assert_that((results[i] == results[i + 1]).all()).is_false()
                        
                    epoch_content = Counter(results[i])
                    indices_gotten = np.array(sorted(list(epoch_content.keys())))
                    assert_that(np.all(np.arange(length) == indices_gotten)).is_true()
                    assert_that(min(epoch_content.values())).is_equal_to(1)
                    assert_that(max(epoch_content.values())).is_less_than_or_equal_to(2)
                else:
                    assert_that(set(results[i])).is_equal_to(set(indices))
                

def test_traversal_sequential_1():
    prep_and_run_test(1, OrderOption.SEQUENTIAL)

def test_traversal_sequential_2():
    prep_and_run_test(2, OrderOption.SEQUENTIAL)

def test_traversal_sequential_3():
    prep_and_run_test(3, OrderOption.SEQUENTIAL)

def test_traversal_sequential_4():
    prep_and_run_test(4, OrderOption.SEQUENTIAL)

def test_traversal_random_1():
    prep_and_run_test(1, OrderOption.RANDOM)

def test_traversal_random_2():
    prep_and_run_test(2, OrderOption.RANDOM)

def test_traversal_random_3():
    prep_and_run_test(3, OrderOption.RANDOM)

def test_traversal_random_4():
    prep_and_run_test(4, OrderOption.RANDOM)

def test_traversal_quasirandom_1():
    prep_and_run_test(1, OrderOption.QUASI_RANDOM)

@pytest.mark.skip()
def test_traversal_quasirandom_2():
    prep_and_run_test(2, OrderOption.QUASI_RANDOM)

@pytest.mark.skip()
def test_traversal_quasirandom_3():
    prep_and_run_test(3, OrderOption.QUASI_RANDOM)

@pytest.mark.skip()
def test_traversal_quasirandom_4():
    prep_and_run_test(4, OrderOption.QUASI_RANDOM)

def test_traversal_sequential_distributed_with_indices():
    prep_and_run_test(2, OrderOption.SEQUENTIAL, True)

def test_traversal_random_distributed_with_indices():
    prep_and_run_test(2, OrderOption.RANDOM, True)

@pytest.mark.skip()
def test_traversal_quasi_random_distributed_with_indices():
    prep_and_run_test(2, OrderOption.QUASI_RANDOM, True)


"""
Random translate
"""
import numpy as np
from numpy.random import randint
from typing import Callable, Optional, Tuple
from dataclasses import replace
from ..pipeline.allocation_query import AllocationQuery
from ..pipeline.operation import Operation
from ..pipeline.state import State
from ..pipeline.compiler import Compiler

class RandomTranslate(Operation):
    """Translate each image randomly in vertical and horizontal directions
    up to specified number of pixels.

    Parameters
    ----------
    padding : int
        Max number of pixels to translate in any direction.
    fill : tuple
        An RGB color ((0, 0, 0) by default) to fill the area outside the shifted image.
    """

    def __init__(self, padding: int, fill: Tuple[int, int, int] = (0, 0, 0)):
        super().__init__()
        self.padding = padding
        self.fill = np.array(fill)

    def generate_code(self) -> Callable:
        my_range = Compiler.get_iterator()
        pad = self.padding
        fill = self.fill

        def translate(images, dst):
            n, h, w, _ = images.shape
            dst[:] = fill
            dst[:, pad:pad+h, pad:pad+w] = images
            for i in my_range(n):
                y_coord = randint(low=0, high=2 * pad + 1)
                x_coord = randint(low=0, high=2 * pad + 1)
                images[i] = dst[i, y_coord:y_coord+h, x_coord:x_coord+w]

            return images

        translate.is_parallel = True
        return translate

    def declare_state_and_memory(self, previous_state: State) -> Tuple[State, Optional[AllocationQuery]]:
        h, w, c = previous_state.shape
        return (replace(previous_state, jit_mode=True), \
                AllocationQuery((h + 2 * self.padding, w + 2 * self.padding, c), previous_state.dtype))




from dataclasses import replace
import torch as ch
from ffcv.pipeline.allocation_query import AllocationQuery
from ffcv.pipeline.compiler import Compiler
import numpy as np
from typing import Callable
from assertpy import assert_that
from torch.utils.data import Dataset
import logging
import os
from assertpy import assert_that
from tempfile import NamedTemporaryFile
from ffcv.pipeline.operation import Operation
from ffcv.transforms.ops import ToTensor
from multiprocessing import cpu_count

from ffcv.writer import DatasetWriter
from ffcv.reader import Reader
from ffcv.loader import Loader
from ffcv.fields import IntField, FloatField, BytesField
from ffcv.fields.basics import FloatDecoder
from ffcv.pipeline.state import State

from test_writer import DummyDataset

numba_logger = logging.getLogger('numba')
numba_logger.setLevel(logging.WARNING)


class Doubler(Operation):

    def generate_code(self) -> Callable:
        def code(x, dst):
            dst[:x.shape[0]] = x * 2
            return dst
        return code

    def declare_state_and_memory(self, previous_state: State):
        return (previous_state, AllocationQuery(previous_state.shape, previous_state.dtype, previous_state.device))

def run_test(bs, exp_length, drop_last=True):
    length = 600
    batch_size = bs
    with NamedTemporaryFile() as handle:
        file_name = handle.name
        dataset = DummyDataset(length)
        writer = DatasetWriter(file_name, {
            'index': IntField(),
            'value': FloatField()
        })

        writer.from_indexed_dataset(dataset)

        Compiler.set_enabled(True)

        loader = Loader(file_name, batch_size, num_workers=min(5, cpu_count()), seed=17,
                drop_last=drop_last,
                        pipelines={
                            'value': [FloatDecoder(), Doubler(), ToTensor()]
                        })

        assert_that(loader).is_length(exp_length)
        another_partial = drop_last
        for (batch, _) in loader:
            if batch.shape[0] != bs:
                assert_that(another_partial).is_false()
                another_partial = True


def test_partial():
    run_test(7, 85, True)

def test_not_partial():
    run_test(7, 86, False)

def test_not_partial_multiple():
    run_test(60, 10, False)

def test_partial_multiple():
    run_test(60, 10, True)


from typing import Sequence

import numpy as np
from torch.utils.data import DistributedSampler

from .base import TraversalOrder

class Random(TraversalOrder):

    def __init__(self, loader:'Loader'):
        super().__init__(loader)

        if self.distributed:
            self.sampler = DistributedSampler(self.indices,
                                              shuffle=True,
                                              seed=self.seed,
                                              drop_last=False)


    def sample_order(self, epoch: int) -> Sequence[int]:
        if not self.distributed:
            generator = np.random.default_rng(self.seed + epoch if self.seed is not None else None)
            return generator.permutation(self.indices)

        self.sampler.set_epoch(epoch)

        return self.indices[np.array(list(self.sampler))]



from typing import Callable, Optional, Tuple
from ..pipeline.allocation_query import AllocationQuery
from ..pipeline.operation import Operation
from ..pipeline.state import State
from dataclasses import replace

class Squeeze(Operation):
    """Remove given dimensions of input of size 1.
    Operates on tensors.

    Parameters
    ----------
    *dims : List[int]
        Dimensions to squeeze.
    """

    def __init__(self, *dims):
        super().__init__()
        self.dims = dims

    def generate_code(self) -> Callable:
        def squeeze(inp, _):
            inp.squeeze_(*self.dims)
            return inp
        return squeeze

    def declare_state_and_memory(self, previous_state: State) -> Tuple[State, Optional[AllocationQuery]]:
        return replace(previous_state, shape=[x for x in previous_state.shape if not x == 1]), None



import argparse

import pandas as pd
from terminaltables import SingleTable

from .suites import *
from .decorator import run_all

parser = argparse.ArgumentParser(description='Run ffcv micro benchmarks')
parser.add_argument('--runs', '-n', type=int,
                    help='Use the median of --runs runs of each test',
                    default=3)
parser.add_argument('--warm-up', '-w', type=int,
                    help='Runs each test --warm-up times before measuring',
                    default=1)
parser.add_argument('--pattern', '-p', type=str,
                    help='Run only tests matching this (glob style) pattern',
                    default='*')
parser.add_argument('--output', '-o', type=str, default=None,
                    help='If defined will write to file instead of stdout.')

args = parser.parse_args()

all_results = run_all(args.runs,
                      args.warm_up,
                      pattern=args.pattern)

result_data = []
for suite_name, results in all_results.items():
    column_names = results[0].keys()
    table_data = [list(column_names)]
    
    for result in results:
        result_data.append({
            'suite_name': suite_name,
            **result
        })
        table_data.append(result.values())
        
    table = SingleTable(table_data, title=suite_name)

    if args.output is None:
        print(table.table)

if args.output is not None:
    frame = pd.DataFrame(result_data)
    frame.to_csv(args.output)



import os
import uuid
import numpy as np
import torch as ch
from torch.utils.data import Dataset
from torchvision import transforms as tvt
from assertpy import assert_that
from tempfile import NamedTemporaryFile
from torchvision.datasets import CIFAR10
from torchvision.utils import save_image, make_grid
from torch.utils.data import Subset
from ffcv.fields.basics import IntDecoder
from ffcv.fields.rgb_image import SimpleRGBImageDecoder

from ffcv.writer import DatasetWriter
from ffcv.fields import IntField, RGBImageField
from ffcv.loader import Loader
from ffcv.pipeline.compiler import Compiler
from ffcv.transforms import *

SAVE_IMAGES = True
IMAGES_TMP_PATH = '/tmp/ffcv_augtest_output'
if SAVE_IMAGES:
    os.makedirs(IMAGES_TMP_PATH, exist_ok=True)

UNAUGMENTED_PIPELINE=[
    SimpleRGBImageDecoder(),
    ToTensor(),
    ToTorchImage()
]

def run_test(length, pipeline, should_compile=False, aug_name=''):
    my_dataset = Subset(CIFAR10(root='/tmp', train=True, download=True), range(length))

    with NamedTemporaryFile() as handle:
        name = handle.name
        writer = DatasetWriter(name, {
            'image': RGBImageField(write_mode='smart', 
                                max_resolution=32),
            'label': IntField(),
        }, num_workers=2)

        writer.from_indexed_dataset(my_dataset, chunksize=10)

        Compiler.set_enabled(should_compile)

        loader = Loader(name, batch_size=7, num_workers=2, pipelines={
            'image': pipeline,
            'label': [IntDecoder(), ToTensor(), Squeeze()]
        },
        drop_last=False)

        unaugmented_loader = Loader(name, batch_size=7, num_workers=2, pipelines={
            'image': UNAUGMENTED_PIPELINE,
            'label': [IntDecoder(), ToTensor(), Squeeze()]
        }, drop_last=False)

        tot_indices = 0
        tot_images = 0
        for it_num, ((images, labels), (original_images, original_labels)) in enumerate(zip(loader, unaugmented_loader)):
            tot_indices += labels.shape[0]
            tot_images += images.shape[0]
            
            for label, original_label in zip(labels, original_labels):
                assert_that(label).is_equal_to(original_label)
            
            if SAVE_IMAGES and it_num == 0:
                save_image(make_grid(ch.concat([images, original_images])/255., images.shape[0]), 
                        os.path.join(IMAGES_TMP_PATH, aug_name + '-' + str(uuid.uuid4()) + '.jpeg'))

        assert_that(tot_indices).is_equal_to(len(my_dataset))
        assert_that(tot_images).is_equal_to(len(my_dataset))

def test_cutout():
    for comp in [True, False]:
        run_test(100, [
            SimpleRGBImageDecoder(),
            Cutout(8),
            ToTensor(),
            ToTorchImage()
        ], comp, 'cutout')


def test_flip():
    for comp in [True, False]:
        run_test(100, [
            SimpleRGBImageDecoder(),
            RandomHorizontalFlip(1.0),
            ToTensor(),
            ToTorchImage()
        ], comp, 'flip')


def test_module_wrapper():
    for comp in [True, False]:
        run_test(100, [
            SimpleRGBImageDecoder(),
            ToTensor(),
            ToTorchImage(),
            ModuleWrapper(tvt.Grayscale(3)),
        ], comp, 'module')


def test_mixup():
    for comp in [True, False]:
        run_test(100, [
            SimpleRGBImageDecoder(),
            ImageMixup(.5, False),
            ToTensor(),
            ToTorchImage()
        ], comp, 'mixup')


def test_poison():
    mask = np.zeros((32, 32, 3))
    # Red sqaure
    mask[:5, :5, 0] = 1
    alpha = np.ones((32, 32))

    for comp in [True, False]:
        run_test(100, [
            SimpleRGBImageDecoder(),
            Poison(mask, alpha, list(range(100))),
            ToTensor(),
            ToTorchImage()
        ], comp, 'poison')

def test_random_resized_crop():
    for comp in [True, False]:
        run_test(100, [
            SimpleRGBImageDecoder(),
            RandomResizedCrop(scale=(0.08, 1.0), 
                            ratio=(0.75, 4/3),
                            size=32),
            ToTensor(),
            ToTorchImage()
        ], comp, 'rrc')


def test_translate():
    for comp in [True, False]:
        run_test(100, [
            SimpleRGBImageDecoder(),
            RandomTranslate(padding=10),
            ToTensor(),
            ToTorchImage()
        ], comp, 'translate')


## Torchvision Transforms
def test_torchvision_greyscale():
    run_test(100, [
        SimpleRGBImageDecoder(),
        ToTensor(),
        ToTorchImage(),
        tvt.Grayscale(3),
        ], aug_name='tv_grey')

def test_torchvision_centercrop_pad():
    run_test(100, [
        SimpleRGBImageDecoder(),
        ToTensor(),
        ToTorchImage(),
        tvt.CenterCrop(10),
        tvt.Pad(11)
        ], aug_name='tv_crop_pad')

def test_torchvision_random_affine():
    run_test(100, [
        SimpleRGBImageDecoder(),
        ToTensor(),
        ToTorchImage(),
        tvt.RandomAffine(25),
        ], aug_name='tv_random_affine')

def test_torchvision_random_crop():
    run_test(100, [
        SimpleRGBImageDecoder(),
        ToTensor(),
        ToTorchImage(),
        tvt.Pad(10),
        tvt.RandomCrop(size=32),
        ], aug_name='tv_randcrop')

def test_torchvision_color_jitter():
    run_test(100, [
        SimpleRGBImageDecoder(),
        ToTensor(),
        ToTorchImage(),
        tvt.ColorJitter(.5, .5, .5, .5)
    ], aug_name='tv_colorjitter')



from collections import defaultdict
from threading import Thread, Event
from queue import Queue, Full
from contextlib import nullcontext
from typing import Sequence, TYPE_CHECKING

import torch as ch

from ..traversal_order.quasi_random import QuasiRandom
from ..utils import chunks
from ..pipeline.compiler import Compiler

if TYPE_CHECKING:
    from .loader import Loader
    
IS_CUDA = ch.cuda.is_available()

QUASIRANDOM_ERROR_MSG = '''Not enough memory; try setting quasi-random ordering
(`OrderOption.QUASI_RANDOM`) in the dataloader constructor's `order` argument.
'''

def select_buffer(buffer, batch_slot, count):
    """Util function to select the relevent subpart of a buffer for a given
    batch_slot and batch size"""
    if buffer is None:
        return None
    if isinstance(buffer, tuple):
        return tuple(select_buffer(x, batch_slot, count) for x in buffer)

    return buffer[batch_slot][:count]


class EpochIterator(Thread):
    def __init__(self, loader: 'Loader', order: Sequence[int]):
        super().__init__(daemon=True)
        self.loader: 'Loader' = loader
        self.order = order
        self.metadata = loader.reader.metadata
        self.current_batch_slot = 0
        batches = list(chunks(order, self.loader.batch_size))
        self.iter_ixes = iter(batches)
        self.closed = False
        self.output_queue = Queue(self.loader.batches_ahead)
        self.terminate_event = Event()
        self.memory_context = self.loader.memory_manager.schedule_epoch(
            batches)

        if IS_CUDA:
            self.current_stream = ch.cuda.current_stream()

        try:
            self.memory_context.__enter__()
        except MemoryError as e:
            if loader.traversal_order != QuasiRandom:
                print(QUASIRANDOM_ERROR_MSG)
                print('Full error below:')

            raise e

        self.storage_state = self.memory_context.state

        self.cuda_streams = [(ch.cuda.Stream() if IS_CUDA else None)
                             for _ in range(self.loader.batches_ahead + 2)]

        self.memory_allocations = self.loader.graph.allocate_memory(
            self.loader.batch_size,
            self.loader.batches_ahead + 2
        )

        self.start()

    def run(self):

        events = [None for _ in self.cuda_streams]

        try:
            b_ix = 0
            Compiler.set_num_threads(self.loader.num_workers)
            while True:
                ixes = next(self.iter_ixes)
                slot = self.current_batch_slot
                self.current_batch_slot = (
                    slot + 1) % (self.loader.batches_ahead + 2)
                result = self.run_pipeline(b_ix, ixes, slot, events[slot])
                # print("RES", b_ix, "ready")
                to_output = (slot, result)
                while True:
                    try:
                        self.output_queue.put(to_output, block=True, timeout=0.5)
                        break
                    except Full:
                        pass

                    if self.terminate_event.is_set():
                        return
                if IS_CUDA:
                    # print("SUB", b_ix)
                    # We were able to submit this batch
                    # Therefore it means that the user must have entered the for loop for
                    # (batch_slot - batch_ahead + 1) % (batches ahead + 2)
                    # Therefore batch_slot - batch_ahead must have all it's work submitted
                    # We will record an event of all the work submitted on the main stream
                    # and make sure no one overwrite the data until they are done
                    just_finished_slot = (slot - self.loader.batches_ahead - 1) % (self.loader.batches_ahead + 2)
                    # print("JFS", just_finished_slot)
                    event = ch.cuda.Event()
                    event.record(self.current_stream)
                    events[just_finished_slot] = event
                b_ix += 1

        except StopIteration:
            self.output_queue.put(None)

    def run_pipeline(self, b_ix, batch_indices, batch_slot, cuda_event):
        self.memory_context.start_batch(b_ix)
        args = []
        if IS_CUDA:
            stream = self.cuda_streams[batch_slot]
            ctx = ch.cuda.stream(stream)
        else:
            ctx = nullcontext()
        first_stage = False


        code, outputs = self.loader.code
        with ctx:
            if IS_CUDA:
                if cuda_event:
                    cuda_event.wait()

            args = {
                'batch_indices': batch_indices,
                'storage_state': self.storage_state,
                'metadata': self.metadata,
                **{
                    f'memory_{k}':select_buffer(v, batch_slot, len(batch_indices))
                    for (k, v) in self.memory_allocations['operation'].items()
                },
                **{
                    f'shared_memory_{k}': select_buffer(v, batch_slot, len(batch_indices))
                    for (k, v) in self.memory_allocations['shared'].items()
                }
            }

            for stage_code, define_outputs in code:
                results = stage_code(**args)
                for node_id, result in zip(define_outputs, results):
                    args[f'result_{node_id}'] = result
                pass

            result = tuple(args[f'result_{x}'] for x in outputs)
            return result

    def __next__(self):
        result = self.output_queue.get()
        if result is None:
            self.close()
            raise StopIteration()
        slot, result = result
        if IS_CUDA:
            stream = self.cuda_streams[slot]
            # We wait for the copy to be done
            self.current_stream.wait_stream(stream)
        return result

    def __iter__(self):
        return self

    def close(self):
        self.terminate_event.set()
        if not self.closed:
            self.memory_context.__exit__(None, None, None)

    def __del__(self):
        self.close()
        



from abc import abstractmethod, ABCMeta, ABC
from contextlib import AbstractContextManager
from collections import defaultdict
from typing import Callable, Mapping, Sequence, Set

import numpy as np
from numba.typed import Dict
from numba import types

from ..reader import Reader
from ..pipeline.compiler import Compiler

class MemoryContext(AbstractContextManager, metaclass=ABCMeta):

    @property
    @abstractmethod
    def state(self):
        raise NotImplementedError()

    @abstractmethod
    def __enter__(self):
        return super().__enter__()

    def start_batch(self, batch:int):
        pass

    @abstractmethod
    def __exit__(self, __exc_type, __exc_value, __traceback):
        return super().__exit__(__exc_type, __exc_value, __traceback)


class MemoryManager(ABC):

    def __init__(self, reader:Reader):
        self.reader = reader
        alloc_table = self.reader.alloc_table

        # Table mapping any address in the file to the size of the data region
        # That was allocated there
        self.ptrs = alloc_table['ptr']
        self.sizes = alloc_table['size']
        order = np.argsort(self.ptrs)
        # Order them so that we can use search sorted
        self.ptrs = self.ptrs[order]
        self.sizes =  self.sizes[order]

        self.ptr_to_size = dict(zip(self.ptrs, self.sizes))

        # We extract the page number by shifting the address corresponding
        # to the page width
        page_size_bit_location = int(np.log2(reader.page_size))
        page_locations = alloc_table['ptr'] >> page_size_bit_location

        sample_to_pages: Mapping[int, Set[int]] = defaultdict(set)
        page_to_samples: Mapping[int, Set[int]] = defaultdict(set)

        # We create a mapping that goes from sample id to the pages it has data
        # Stored to
        # (And the same for the other way around)
        for sid, pid in zip(alloc_table['sample_id'], page_locations):
            sample_to_pages[sid].add(pid)
            page_to_samples[pid].add(sid)

        self.sample_to_pages = sample_to_pages
        self.page_to_samples = page_to_samples

        super().__init__()

    @abstractmethod
    def schedule_epoch(self, batches: Sequence[Sequence[int]]) -> MemoryContext:
        raise NotImplementedError()

    @abstractmethod
    def compile_reader(self, address, size) -> Callable:
        raise NotImplemented()

    @property
    @abstractmethod
    def state_type(self):
        raise NotImplementedError()





"""
Poison images by adding a mask
"""
from typing import Tuple
from dataclasses import replace

import numpy as np
from typing import Callable, Optional, Tuple
from ..pipeline.allocation_query import AllocationQuery
from ..pipeline.operation import Operation
from ..pipeline.state import State
from ..pipeline.compiler import Compiler

class Poison(Operation):
    """Poison specified images by adding a mask with given opacity.
    Operates on raw arrays (not tensors).

    Parameters
    ----------
    mask : ndarray
        The mask to apply to each image.
    alpha: float
        The opacity of the mask.
    indices : Sequence[int]
        The indices of images that should have the mask applied.
    clamp : Tuple[int, int]
        Clamps the final pixel values between these two values (default: (0, 255)).
    """

    def __init__(self, mask: np.ndarray, alpha: np.ndarray,
                 indices, clamp = (0, 255)):
        super().__init__()
        self.mask = mask
        self.indices = np.sort(indices)
        self.clamp = clamp
        self.alpha = alpha

    def generate_code(self) -> Callable:

        alpha = np.repeat(self.alpha[:, :, None], 3, axis=2)
        mask = self.mask.astype('float') * alpha
        to_poison = self.indices
        clamp = self.clamp
        my_range = Compiler.get_iterator()

        def poison(images, temp_array, indices):
            for i in my_range(images.shape[0]):
                sample_ix = indices[i]
                # We check if the index is in the list of indices
                # to poison
                position = np.searchsorted(to_poison, sample_ix)
                if position < len(to_poison) and to_poison[position] == sample_ix:
                    temp = temp_array[i]
                    temp[:] = images[i]
                    temp *= 1 - alpha
                    temp += mask
                    np.clip(temp, clamp[0], clamp[1], out=temp)
                    images[i] = temp
            return images

        poison.is_parallel = True
        poison.with_indices = True

        return poison

    def declare_state_and_memory(self, previous_state: State) -> Tuple[State, Optional[AllocationQuery]]:
        # We do everything in place
        return (replace(previous_state, jit_mode=True), \
                AllocationQuery(shape=previous_state.shape, dtype=np.dtype('float32')))



import os
from tempfile import NamedTemporaryFile
from time import sleep, time
import os, psutil


import numpy as np
import pytest
from tqdm import tqdm
from assertpy import assert_that
from torch.utils.data import Dataset

from ffcv.writer import DatasetWriter
from ffcv.reader import Reader
from ffcv.fields import BytesField, IntField
from ffcv.pipeline.compiler import Compiler
from ffcv import Loader

class DummyDataset(Dataset):

    def __init__(self, l, size):
        self.l = l
        self.size = size

    def __len__(self):
        return self.l

    def __getitem__(self, index):
        if index > self.l:
            raise IndexError
        np.random.seed(index)
        return index, np.random.randint(0, 255, size=self.size, dtype='u1')

def create_and_run(num_samples, size_bytes):
    handle = NamedTemporaryFile()
    with handle:
        name = handle.name
        dataset = DummyDataset(num_samples, size_bytes)
        writer = DatasetWriter(num_samples, name, {
            'index': IntField(),
            'value': BytesField()
        })
        
        Compiler.set_enabled(True)

        with writer:
            writer.write_pytorch_dataset(dataset, num_workers=-1, chunksize=100)
        total_dataset_size = num_samples * size_bytes
        # Dataset should not be in RAM
        process = psutil.Process(os.getpid())
        assert_that(process.memory_info().rss).is_less_than(total_dataset_size)
        
        loader = Loader(name, 128, 10)
        for _ in tqdm(loader):
            assert_that(process.memory_info().rss).is_less_than(total_dataset_size)



@pytest.mark.skipif(bool(os.environ.get('FFCV_RUN_MEMORY_LEAK_TEST', "0")),
                    reason="set FFCV_RUN_MEMORY_LEAK_TEST to enable it")
def test_memory_leak_write():
    create_and_run(128100, 500*300*3)


import torch as ch

from typing import List, Union
from .operation import Operation
from ..transforms.module import ModuleWrapper
from ..transforms import ToTensor

class PipelineSpec:

    def __init__(self, source: Union[str, Operation], decoder: Operation = None,
                 transforms:List[Operation] = None ):

        self.source = source
        self.decoder = decoder
        if transforms is None:
            transforms = []
        self.transforms = transforms
        self.default_pipeline = (decoder is None
                                 and not transforms
                                 and isinstance(source, str))

    def __repr__(self):
        return repr((self.source, self.decoder, self.transforms))

    def __str__(self):
        return self.__repr__()

    def accept_decoder(self, Decoder, output_name):
        if not isinstance(self.source, str) and self.decoder is not None:
            raise ValueError("Source can't be a node and also have a decoder")

        if Decoder is not None:
            # The first element of the operations is a decoder
            if self.transforms and isinstance(self.transforms[0], Decoder):
                self.decoder = self.transforms.pop(0)

            elif self.decoder is None:
                try:
                    self.decoder = Decoder()
                except Exception:
                    msg = f"Impossible to use default decoder for {output_name},"
                    msg += "make sure you specify one in your pipeline."
                    raise ValueError(msg)

        if self.default_pipeline:
            self.transforms.append(ToTensor())

        for i, op in enumerate(self.transforms):
            if isinstance(op, ch.nn.Module):
                self.transforms[i] = ModuleWrapper(op)
            


import ctypes
from numba import njit
import numpy as np
import platform
from ctypes import CDLL, c_int64, c_uint8, c_uint64, POINTER, c_void_p, c_uint32, c_bool, cdll
import ffcv._libffcv

lib = CDLL(ffcv._libffcv.__file__)
if platform.system() == "Windows":
    libc = cdll.msvcrt
    read_c = libc._read
else:
    libc = cdll.LoadLibrary('libc.so.6')
    read_c = libc.pread

read_c.argtypes = [c_uint32, c_void_p, c_uint64, c_uint64]

def read(fileno:int, destination:np.ndarray, offset:int):
    return read_c(fileno, destination.ctypes.data, destination.size, offset)


ctypes_resize = lib.resize
ctypes_resize.argtypes = 11 * [c_int64]

def resize_crop(source, start_row, end_row, start_col, end_col, destination):
    ctypes_resize(0,
                  source.ctypes.data,
                  source.shape[0], source.shape[1],
                  start_row, end_row, start_col, end_col,
                  destination.ctypes.data,
                  destination.shape[0], destination.shape[1])

# Extract and define the interface of imdeocde
ctypes_imdecode = lib.imdecode
ctypes_imdecode.argtypes = [
    c_void_p, c_uint64, c_uint32, c_uint32, c_void_p, c_uint32, c_uint32,
    c_uint32, c_uint32, c_uint32, c_uint32, c_bool, c_bool
]

def imdecode(source: np.ndarray, dst: np.ndarray,
             source_height: int, source_width: int,
             crop_height=None, crop_width=None,
             offset_x=0, offset_y=0, scale_factor_num=1, scale_factor_denom=1,
             enable_crop=False, do_flip=False):
    return ctypes_imdecode(source.ctypes.data, source.size,
                           source_height, source_width, dst.ctypes.data,
                           crop_height, crop_width, offset_x, offset_y, scale_factor_num, scale_factor_denom,
                           enable_crop, do_flip)


ctypes_memcopy = lib.my_memcpy
ctypes_memcopy.argtypes = [c_void_p, c_void_p, c_uint64]

def memcpy(source: np.ndarray, dest: np.ndarray):
    return ctypes_memcopy(source.ctypes.data, dest.ctypes.data, source.size*source.itemsize)




from argparse import ArgumentParser
from typing import List
import time
import numpy as np
from tqdm import tqdm

import torch as ch
import torchvision

from fastargs import get_current_config
from fastargs.decorators import param
from fastargs import Param, Section
from fastargs.validation import And, OneOf

from ffcv.writer import DatasetWriter
from ffcv.fields import IntField, RGBImageField

Section('data', 'arguments to give the writer').params(
    train_dataset=Param(str, 'Where to write the new dataset', required=True),
    val_dataset=Param(str, 'Where to write the new dataset', required=True),
)

@param('data.train_dataset')
@param('data.val_dataset')
def main(train_dataset, val_dataset):
    datasets = {
        'train': torchvision.datasets.CIFAR10('/tmp', train=True, download=True),
        'test': torchvision.datasets.CIFAR10('/tmp', train=False, download=True)
        }

    for (name, ds) in datasets.items():
        path = train_dataset if name == 'train' else val_dataset
        writer = DatasetWriter(path, {
            'image': RGBImageField(),
            'label': IntField()
        })
        writer.from_indexed_dataset(ds)


if __name__ == "__main__":
    config = get_current_config()
    parser = ArgumentParser(description='Fast CIFAR-10 training')
    config.augment_argparse(parser)
    config.collect_argparse_args(parser)
    config.validate(mode='stderr')
    config.summary()

    main()


from .base import MemoryManager, MemoryContext
from .process_cache import ProcessCacheManager
from .os_cache import OSCacheManager

__all__ = ['OSCacheManager', 'ProcessCacheManager',
           'MemoryManager', 'MemoryContext']



from dataclasses import replace
import torch as ch
from ffcv.pipeline.allocation_query import AllocationQuery
from ffcv.pipeline.compiler import Compiler
import numpy as np
from typing import Callable
from assertpy import assert_that
from torch.utils.data import Dataset
import logging
import os
from assertpy import assert_that
from tempfile import NamedTemporaryFile
from ffcv.pipeline.operation import Operation
from ffcv.transforms.ops import ToTensor
from multiprocessing import cpu_count

from ffcv.writer import DatasetWriter
from ffcv.reader import Reader
from ffcv.loader import Loader
from ffcv.fields import IntField, FloatField, BytesField
from ffcv.fields.basics import FloatDecoder
from ffcv.pipeline.state import State

from test_writer import DummyDataset

numba_logger = logging.getLogger('numba')
numba_logger.setLevel(logging.WARNING)


class Doubler(Operation):

    def generate_code(self) -> Callable:
        def code(x, dst):
            dst[:] = x * 2
            return dst
        return code

    def declare_state_and_memory(self, previous_state: State):
        return (previous_state, AllocationQuery(previous_state.shape, previous_state.dtype, previous_state.device))

def test_basic_simple():
    length = 600
    batch_size = 8
    with NamedTemporaryFile() as handle:
        file_name = handle.name
        dataset = DummyDataset(length)
        writer = DatasetWriter(file_name, {
            'index': IntField(),
            'value': FloatField()
        })

        writer.from_indexed_dataset(dataset)

        Compiler.set_enabled(True)

        loader = Loader(file_name, batch_size, num_workers=min(5, cpu_count()), seed=17,
                        pipelines={
                            'value': [FloatDecoder(), Doubler(), ToTensor()],
                        })

        def cond(value):
            value = value[0]
            result = value < 1 and value >= 0.5
            return result

        filtered = loader.filter('value', cond)

        assert_that(len(filtered)).is_greater_than(0)
        for index, values in filtered:
            assert_that(values.shape[0]).is_equal_to(batch_size)
            assert_that(((values < 1) & (values >= 0.5)).all()).is_true()


'''
Random color operations similar to torchvision.transforms.ColorJitter except not supporting hue
Reference : https://github.com/pytorch/vision/blob/main/torchvision/transforms/functional_tensor.py
'''

import numpy as np

from dataclasses import replace
from ..pipeline.allocation_query import AllocationQuery
from ..pipeline.operation import Operation
from ..pipeline.state import State
from ..pipeline.compiler import Compiler



class RandomBrightness(Operation):
    '''
    Randomly adjust image brightness. Operates on raw arrays (not tensors).

    Parameters
    ----------
    magnitude : float
        randomly choose brightness enhancement factor on [max(0, 1-magnitude), 1+magnitude]
    p : float
        probability to apply brightness
    '''
    def __init__(self, magnitude: float, p=0.5):
        super().__init__()
        self.p = p
        self.magnitude = magnitude

    def generate_code(self):
        my_range = Compiler.get_iterator()
        p = self.p
        magnitude = self.magnitude

        def brightness(images, *_):
            def blend(img1, img2, ratio): return (ratio*img1 + (1-ratio)*img2).clip(0, 255).astype(img1.dtype)

            apply_bright = np.random.rand(images.shape[0]) < p
            magnitudes = np.random.uniform(max(0, 1-magnitude), 1+magnitude, images.shape[0])
            for i in my_range(images.shape[0]):
                if apply_bright[i]:
                    images[i] = blend(images[i], 0, magnitudes[i])

            return images

        brightness.is_parallel = True
        return brightness

    def declare_state_and_memory(self, previous_state):
        return (replace(previous_state, jit_mode=True), AllocationQuery(previous_state.shape, previous_state.dtype))



class RandomContrast(Operation):
    '''
    Randomly adjust image contrast. Operates on raw arrays (not tensors).

    Parameters
    ----------
    magnitude : float
        randomly choose contrast enhancement factor on [max(0, 1-magnitude), 1+magnitude]
    p : float
        probability to apply contrast
    '''
    def __init__(self, magnitude, p=0.5):
        super().__init__()
        self.p = p
        self.magnitude = magnitude

    def generate_code(self):
        my_range = Compiler.get_iterator()
        p = self.p
        magnitude = self.magnitude

        def contrast(images, *_):
            def blend(img1, img2, ratio): return (ratio*img1 + (1-ratio)*img2).clip(0, 255).astype(img1.dtype)

            apply_contrast = np.random.rand(images.shape[0]) < p
            magnitudes = np.random.uniform(max(0, 1-magnitude), 1+magnitude, images.shape[0])
            for i in my_range(images.shape[0]):
                if apply_contrast[i]:
                    r, g, b = images[i,:,:,0], images[i,:,:,1], images[i,:,:,2]
                    l_img = (0.2989 * r + 0.587 * g + 0.114 * b).astype(images[i].dtype)
                    images[i] = blend(images[i], l_img.mean(), magnitudes[i])

            return images

        contrast.is_parallel = True
        return contrast

    def declare_state_and_memory(self, previous_state):
        return (replace(previous_state, jit_mode=True), AllocationQuery(previous_state.shape, previous_state.dtype))



class RandomSaturation(Operation):
    '''
    Randomly adjust image color balance. Operates on raw arrays (not tensors).

    Parameters
    ----------
    magnitude : float
        randomly choose color balance enhancement factor on [max(0, 1-magnitude), 1+magnitude]
    p : float
        probability to apply saturation
    '''
    def __init__(self, magnitude, p=0.5):
        super().__init__()
        self.p = p
        self.magnitude = magnitude

    def generate_code(self):
        my_range = Compiler.get_iterator()
        p = self.p
        magnitude = self.magnitude

        def saturation(images, *_):
            def blend(img1, img2, ratio): return (ratio*img1 + (1-ratio)*img2).clip(0, 255).astype(img1.dtype)

            apply_saturation = np.random.rand(images.shape[0]) < p
            magnitudes = np.random.uniform(max(0, 1-magnitude), 1+magnitude, images.shape[0])
            for i in my_range(images.shape[0]):
                if apply_saturation[i]:
                    r, g, b = images[i,:,:,0], images[i,:,:,1], images[i,:,:,2]
                    l_img = (0.2989 * r + 0.587 * g + 0.114 * b).astype(images[i].dtype)
                    l_img3 = np.zeros_like(images[i])
                    for j in my_range(images[i].shape[-1]):
                        l_img3[:,:,j] = l_img
                    images[i] = blend(images[i], l_img3, magnitudes[i])

            return images

        saturation.is_parallel = True
        return saturation

    def declare_state_and_memory(self, previous_state):
        return (replace(previous_state, jit_mode=True), AllocationQuery(previous_state.shape, previous_state.dtype))



from tempfile import NamedTemporaryFile
import torch as ch
from tqdm import tqdm
import time
from multiprocessing import cpu_count

from assertpy import assert_that
import numpy as np
from torch.utils.data import Dataset
from ffcv import DatasetWriter
from ffcv.fields import IntField, NDArrayField
from ffcv import Loader
from ffcv.fields.basics import IntDecoder
from ffcv.fields.ndarray import NDArrayDecoder
from ffcv.loader.loader import OrderOption
from ffcv.transforms import ToDevice, ToTensor, Squeeze
import time

BATCH = 256
SIZE = 25_000
WORKERS = min(10, cpu_count())

class DummyArrayDataset(Dataset):
    def __init__(self, n_samples, shape):
        self.n_samples = n_samples
        self.shape = shape

    def __len__(self):
        return self.n_samples

    def __getitem__(self, index):
        if index >= self.n_samples:
            raise IndexError()
        np.random.seed(index)
        return (np.random.rand(SIZE) > 0.5).astype('bool'), np.random.rand(SIZE).astype('float32'), index

def run_experiment_cuda(weight, loader, sync=False):
    total = 0.
    X = ch.empty(BATCH, SIZE, device=weight.device)
    for X_bool, _, __ in tqdm(loader):
        if sync: ch.cuda.synchronize()
        X.copy_(X_bool)
        total += X @ weight
        total += X @ weight
        total += X @ weight

    return total.sum(0)

def run_cuda(weight, sync):
    n_samples, shape = (BATCH * WORKERS, (SIZE,))
    with NamedTemporaryFile() as handle:
        name = handle.name
        dataset = DummyArrayDataset(n_samples, shape)
        writer = DatasetWriter(name, {
            'mask': NDArrayField(dtype=np.dtype('bool'), shape=(SIZE,)),
            'targets': NDArrayField(dtype=np.dtype('float32'), shape=(SIZE,)),
            'idx': IntField()
        })

        writer.from_indexed_dataset(dataset)

        loader = Loader(
                name,
                batch_size=BATCH,
                num_workers=WORKERS,
                order=OrderOption.QUASI_RANDOM,
                indices=np.arange(n_samples),
                drop_last=False,
                os_cache=True,
                pipelines={
                    'mask': [NDArrayDecoder(), ToTensor(), ToDevice(ch.device('cuda:0'), non_blocking=False)],
                    'targets': [NDArrayDecoder(), ToTensor(), ToDevice(ch.device('cuda:0'), non_blocking=False)],
                    'idx': [IntDecoder(), ToTensor(), Squeeze(), ToDevice(ch.device('cuda:0'), non_blocking=False)]
                })

        return run_experiment_cuda(weight, loader, sync)

def test_cuda():
    weight = ch.randn(SIZE, SIZE).cuda()
    async_1 = run_cuda(weight, False)
    sync_1 = run_cuda(weight, True)
    sync_2 = run_cuda(weight, True)
    print(async_1)
    print(sync_1)
    print(sync_2)
    print(ch.abs(sync_1 - sync_2).max())
    print(ch.abs(sync_1 - async_1).max())
    assert ch.abs(sync_1 - sync_2).max().cpu().item() < float(WORKERS), 'Sync-sync mismatch'
    assert ch.abs(async_1 - sync_1).max().cpu().item() < float(WORKERS), 'Async-sync mismatch'

# test_cuda()


from collections import defaultdict
from dataclasses import dataclass
from typing import Mapping
from queue import Queue

import numpy as np

from .page_reader import PageReader


@dataclass
class Schedule:
    # Number of slots needed
    num_slots: int
    # Which slot to use for each page
    page_to_slot: Mapping[int, int]
    # First iteration a page can be loaded
    can_prefetch_at: Mapping[int, int]
    # Iteration at which a page *has* to be loaded
    entering_at: Mapping[int, int]
    # Iteration at which we can discard a page
    leaving_at: Mapping[int, int]

def compute_schedule(pages_in_batch, prefetch_ahead = 3):
    # We determine what is the early and latest times we will need a page
    page_end = {}
    page_start = {}
    for b_id, pages in enumerate(pages_in_batch):
        for page in pages:
            page_end[page] = b_id
            if page not in page_start:
                page_start[page] = b_id

    # We determine which pages are
    # - Can be preloaded
    # - Are needed
    # - Can be diposed of
    # At a given batch
    entering_at = defaultdict(set)
    can_prefetch_at = defaultdict(set)
    leaving_at = defaultdict(set)
    for page in page_start.keys():
        prefetch_start = max(0, page_start[page] - prefetch_ahead)
        can_prefetch_at[prefetch_start].add(page)
        entering_at[page_start[page]].add(page)
        leaving_at[page_end[page] + 1].add(page)


     # We now find how many pages we need to keep in our buffer
     # We also determine where which page is going to reside
    next_slot = 0
    page_to_slot = {}
    free_slots = set()

    # For each batch
    for b_id in range(len(pages_in_batch)):
        # First we free the pages that are leaving
        for page in leaving_at[b_id]:
            free_slots.add(page_to_slot[page])

        # We use the prefetch timing here because we want to be able
        # To start prefetching ahead of time and not overwrite a slot
        # That is currently used
        for page in can_prefetch_at[b_id]:
            # Then we find a slot for the incoming pages
            if free_slots:
                # There is a slot available for this page
                slot = free_slots.pop()
            else:
                # We have to allocate a new slot because we ran out
                slot = next_slot
                next_slot += 1

            page_to_slot[page] = slot

    return Schedule(next_slot, page_to_slot,
                    can_prefetch_at, entering_at, leaving_at)

class ScheduleExecutor():

    def __init__(self, fname: str, schedule: Schedule,
                 memory: np.ndarray, num_workers: int=12):
        self.fname = fname
        self.schedule = schedule
        self.memory = memory
        self.queries = Queue()
        self.loaded_queue = Queue()
        self.num_workers = num_workers
        self.entered = False
        self.next_batch = 0
        self.loaded = set()

    def __enter__(self):
        msg = "You can only enter a ScheduleExecutor once"
        if self.entered:
            raise Exception(msg)
        self.entered = True
# Create the number of threads we were asked to
        threads = []
        for _ in range(self.num_workers):
            thread = PageReader(self.fname, self.queries,
                                self.loaded_queue, self.memory)
            thread.start()
            threads.append(thread)

        self.threads = threads

    def __exit__(self, *_):
        # Terminating the child threads
        for _ in range(self.num_workers):
            self.queries.put(None)

    def load_batch(self, current_batch):
        assert current_batch == self.next_batch

        # Start prefetching everything we are allowed to
        to_prefetch = self.schedule.can_prefetch_at[current_batch]
        for page_to_fetch in to_prefetch:
            q = (page_to_fetch, self.schedule.page_to_slot[page_to_fetch])
            self.queries.put(q)

        # Wait until we have all the pages we need
        to_wait_for = self.schedule.entering_at[current_batch]
        for page in to_wait_for:
            while page not in self.loaded:
                next_loaded = self.loaded_queue.get()
                self.loaded.add(next_loaded)

        # We enforce that we read in order otherwise our
        # assumptions are broken
        self.next_batch = current_batch + 1
